/*
Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* System includes */
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/types.h>

/* Our includes */
#include "flow_stats.h"
#include "conn_stats.h"
#include "rl_stats.h"
#include "process.h"
#include "dward.config"
#include "dward.h"
#include "gui.h"
#include "functions.h"
#include "replay.h"

extern struct hash_table* ConnHash; 
int sent = 1;
int ProcessedMessages = 0;
long classifications = 0;

/*********** PREFIX CONFIG RELATED ***********/
__u32 LOCAL_ADDRESS = 0, WATCHDOG_ADDRESS;
int LOCAL_MASK = -1, WATCHDOG_PORT, GUIPORT;
char MY_ADDRESS[17];

/*********** DNS SERVERS **************/
__u32 DNS_servers[10];
int num_dns;
FILE* dnsfile = 0;

/*********** DEBUG RELATED ***********/
/* If this flag is set from command line
   then debug messages are written to files 
   in debug directory, for all
   important functions */
int DEBUG = FALSE;
/* Debug files */
extern FILE* flowdebug;
extern FILE* conndebug;
extern FILE* mondebug;
extern FILE* classdebug;
extern FILE* rldebug;
/* Log absolute time, if FALSE relative time will be
   logged from the start of the D-WARD run */
int ABSOLUTE_TIME = FALSE;
/* Apply relaxed models to traffic from this address */
__u32 aggressive_address[MAX_TCP_AGGRESSIVE];
int num_aggressive=0;
/* File for monitoring traffic from a certain source address */
FILE* observed_stats = 0;

/*********** TRACE REPLAY RELATED ***********/
/* This flag indicates that trace data should be used for
   generation of background traffic */
int REPLAY = FALSE;
char tracefile_name[80];
FILE* tracefile;
void get_trace();
void read_trace();

/*********** GUI RELATED ***********/
/* This flag indicates that data should be generated for 
   GUI animation */
int GUI = FALSE;
int GUI_SERVER = FALSE;
int gui_addr = 0;
extern struct hash_table* FlowHash; 
extern int pending_flow;
/* Socket for communication with GUI */
int guifd;
/* Victim addresses */
char victims[100][80];
int num_victims = 0;
long timenow, bt;
int gput, fwp, fwb, old_bad = 0, old_good = 0;
FILE* guidebug = NULL;
int gdtot, bdtot;

/* Function that reports relevant statistics to GUI */
void send_to_GUI(int sent_to, int received_from, int sent_bytes, 
		 int received_bytes, 
		 int attack_packets, int attack_bytes, 
		 unsigned int address)
{
    unsigned char buf[10000];
    int k, rl = 0;
    for (k=0;k<num_victims;k++)
      {
	if (address == inet_addr(victims[k]))
	  {
	    memset(buf,20,100);
	    rl = find_rate_limit_num(address);
	    sprintf(buf,"2 %ld %d %d %d %d %d %d %d %d %d %s %s,",
		    timenow,rl,sent_to,received_from,
		    sent_bytes,received_bytes,
		    attack_packets,attack_bytes,gdtot,bdtot,
		    MY_ADDRESS,victims[k]);
	    send(guifd,buf,100,0);
	  }
      }
}

/*********** RED TEAM RELATED ***********/
/* If this flag is set from command line
   Red Team statistics are written to a file 
   in a directory given in command line */
int RED_TEAM = FALSE;
/* Statisctics files */
char statsdir[80];
FILE* dward_coss_stats = NULL;
FILE* coss_stats = NULL;
FILE* dward_stats = NULL;
FILE* dward_stats_parsable = NULL;

/*********** DefCOM  RELATED ***********/
int DefCOM = FALSE;

/* Socket on which to listen */
int defsock = -1;

void parse_defcom_message(char* message)
{
  int count=0;
  char temp[80];
  get_word(message, &count, temp);
  if (atoi(temp) == 1)
    {
      int index;
      /* Alert message */
      eat_space(message,&count);
      get_word(message, &count, temp);
      /* Ignore sender, get victim */
      eat_space(message,&count);
      get_word(message, &count, temp);
      index=find_flow(temp);
      if (index != NOT_FOUND)
	  set_alert(inet_addr(temp));
    }
  else if (atoi(temp) == 2)
    {
      int index;
      char address[80];
      /* Rate limit message */
      eat_space(message,&count);
      get_word(message, &count, temp);
      /* Ignore sender, get victim */
      eat_space(message,&count);
      get_word(message, &count, address);      
      index=find_flow(address);
      if (index != NOT_FOUND)
	{
	  /* Get rate limit */
	  eat_space(message,&count);
	  get_word(message, &count, temp);
	  set_rate_limit(inet_addr(address), atoi(temp));
	}
    }
  return;
}

void receive_defcom()
{
  const int BUF_SIZE = 500;
  struct sockaddr_in their_addr;    
  fd_set readfds;
  char buf[BUF_SIZE];
  int addr_len = sizeof(struct sockaddr_in), n;
  
  FD_CLR(defsock, &readfds);
  FD_SET(defsock, &readfds);
  while (TRUE)
    {
      if (select(defsock+1, &readfds, NULL, NULL, NULL) == -1) 
	{
	  perror("select");
	  exit(1);
	}
      if (FD_ISSET(defsock,&readfds))
	{
	  n=recvfrom(defsock,buf, BUF_SIZE-1, 0,
		     (struct sockaddr *)&their_addr,&addr_len);
	  buf[n]=0;
	  parse_defcom_message(buf);
	}
    }
} 


/*********** COSSACK RELATED ***********/
/* This flag indicates if there is communication 
   between D-WARD entities */
int COSSACK = FALSE;
int numAttacks = 1;
/* Socket for communication with multicast group */
int commfd;

/* Send multicas message to the group */
void sendMulticastMessage(char* target, int type)
{
  fd_set writefds;
  char buf[100];
  return;
  FD_CLR(commfd, &writefds);
  FD_SET(commfd, &writefds);

  if (select(commfd+1, NULL, &writefds, NULL, NULL) == -1) 
    {
      perror("select");
      exit(1);
    }
  if (FD_ISSET(commfd,&writefds))
    {
      memset(buf,20,100);
      sprintf(buf,"Dward Key=%d; %s %d\n",numAttacks++,target,type);
      send(commfd,buf,100,0);
      memset(buf,0,100);
      sprintf(buf,"5 %s,",MY_ADDRESS);
      send(guifd,buf,100,0);
    }
}

/* Process multicast message */
void process_multicast(char* message)
{
  enum message_field {DSTS=0, SRCS, PROTO, UNITS, STATUS, LIMIT};
  int i, tj, j, curport,curmask, protocol,status,limit,k;
  int start = FALSE, num_dsts = 0, srcs = 0;
  unsigned int curaddress = 0, isAttack;
  enum message_field curfield = DSTS;
  char temp[80];
  unsigned int dsts[NUM_DSTS];
  /* Parse the message */
  for(i=0;i<strlen(message);i++)
    {
      if (message[i]=='=' && (curfield > SRCS))
	{
	  start = TRUE;
	  tj = 0;
	}
      else if (message[i]=='[' && (curfield <= SRCS))
	{
	  start = TRUE;
	  tj = 0;
	}
      else if (message[i]==';')
	{
	  start = FALSE;
	  temp[tj] = 0;
	  if (curfield == PROTO)
	    protocol = atoi(temp);
	  else if (curfield == STATUS)
	    {
	      if (!strcmp(temp,"Stopped"))
		status = 0;
	      else 
		status = 1;
	    }
	  if (curfield == LIMIT)
	    limit = atoi(temp)/8.0;
	  curfield++;
	}
      else if (message[i] == '/' && curfield <= SRCS)
	{
	  temp[tj] = 0;
	  curaddress = ntohl(inet_addr(temp));
	  tj = 0;
	}
      else if (message[i] == ':' && curfield <= SRCS)
	{
	  temp[tj] = 0;
	  curmask = atoi(temp);
	  tj = 0;
	}
      else if (message[i] == '{' && curfield <= SRCS)
	{
	  temp[tj] = 0;
	  curport = atoi(temp);
	  tj = 0;
	  /* Now check if this address/mask is relevant for D-WARD */
	  if ((curfield == DSTS) && curaddress)
	    {
	      int mask = (~0) << (32-curmask);
	      /* It is destination address, check if it is in hash */
	      for(j=0;j<FlowHash->size;j++)
		{
		  /* For each valid flow */
		  if(FlowHash->entries[j].state == VALID)
		    {
		      unsigned int address = ntohl(*((unsigned int*)
						     FlowHash->entries[j].key));
		      if ((address & mask) == (curaddress & mask))
			{
			  for(k=0;k<num_dsts;k++)
			    if (dsts[k]==address)
			      break;
			  if(k==num_dsts)
			    dsts[num_dsts++] = address;
			}
		    }
		}
	    }
	  else
	    {
	      unsigned int address = ntohl(LOCAL_ADDRESS);
	      int mask;
	      if (curmask < (LOCAL_MASK))
		mask = (~0) << (32-curmask);
	      else
		mask = (~0) << (32-LOCAL_MASK);
	      if ((address & mask) == (curaddress & mask))
		srcs ++;
	    }
	  start = FALSE;
	}
      else if (message[i] == ',' && curfield <= SRCS)
	{
	  start = TRUE;
	  tj = 0;
	}
      else if (start && message[i]!=' ')
	temp[tj++]=message[i];
    }
  if (num_dsts && srcs)
    {
      if (status)
	{
	  if (protocol == TCP)
	    isAttack = TCP_ATTACK;
	  else if (protocol == UDP)
	    isAttack = UDP_ATTACK;
	  else if (protocol == ICMP)
	    isAttack = ICMP_ATTACK;
	  else
	    isAttack = OTHER_ATTACK;
	}
      for(j=0;j<num_dsts;j++)
	{
	  int flag=
	    set_cossack_flag_and_rate(htonl(dsts[j]), status, limit, isAttack);
	  if (RED_TEAM && flag)
	    {
	      struct in_addr in;  
	      in.s_addr = htonl(dsts[j]);
	      if (status)
		{
		  fprintf(dward_coss_stats, 
			  "\t\t--> We are participating in ");
		  switch(isAttack)
		    {
		    case UDP_ATTACK:
		      fprintf(dward_coss_stats,"UDP ");
		      break;
		    case TCP_ATTACK:
		      fprintf(dward_coss_stats,"TCP ");
		      break;
		    case ICMP_ATTACK:
		      fprintf(dward_coss_stats,"ICMP ");
		      break;
		    case OTHER_ATTACK:
		      fprintf(dward_coss_stats,"OTHER ");
		      break;
		    }
		  fprintf(dward_coss_stats," attack on %s, ",
			  inet_ntoa(in));
		  if ((isAttack==UDP_ATTACK) || (isAttack == OTHER_ATTACK))
		    fprintf(dward_coss_stats,"will install rate limit %d\n",
			    limit);
		  else
		    fprintf(dward_coss_stats,"will determine appropriate "
			    "rate limit \n");
		}
	      else
		{
		  switch(isAttack)
		    {
		    case UDP_ATTACK:
		      fprintf(dward_coss_stats,"/t/t--> UDP ");
		      break;
		    case TCP_ATTACK:
		      fprintf(dward_coss_stats,"/t/t--> TCP ");
		      break;
		    case ICMP_ATTACK:
		      fprintf(dward_coss_stats,"/t/t--> ICMP ");
		      break;
		    case OTHER_ATTACK:
		      fprintf(dward_coss_stats,"/t/t--> OTHER ");
		      break;
		    }
		  fprintf(dward_coss_stats," attack on %s has stopped, ",
			  inet_ntoa(in));
		  if ((isAttack==UDP_ATTACK) || (isAttack == OTHER_ATTACK))
		    fprintf(dward_coss_stats,"will install rate limit %d\n",
			    limit);
		  else
		    fprintf(dward_coss_stats,"will determine appropriate "
			    "rate limit \n");	      
		}
	      fflush(dward_coss_stats);
	    }
	}
    }
}

/* Send and receive data from multicast group */
void multicast()
{
  const int BUF_SIZE = 500;
  fd_set readfds, writefds;
  char buf[BUF_SIZE];

  FD_CLR(commfd, &readfds);
  FD_SET(commfd, &readfds);
  FD_CLR(commfd, &writefds);
  FD_SET(commfd, &writefds);
  while (TRUE)
    {
      if (select(commfd+1, &readfds, NULL, NULL, NULL) == -1) 
	{
	  perror("select");
	  exit(1);
	}
      if (FD_ISSET(commfd,&readfds))
	{
	  int numbytes = recv(commfd, buf, BUF_SIZE, 0);
	  buf[numbytes] = 0;
	  if (RED_TEAM && numbytes)
	    {
	      if (ABSOLUTE_TIME)
		fprintf(dward_coss_stats,"%d: \tD-WARD received COSSACK"
			" message: %s",(int)time(0), buf);
	      else
		fprintf(dward_coss_stats,"%d: D-WARD received COSSACK"
			" message: %s", (int)timenow, buf);
	      fflush(dward_coss_stats);
	    }
	  process_multicast(buf);
	  FD_CLR(commfd, &readfds);
	  FD_SET(commfd, &readfds);
	}
    }
}

/*********** CLASSIFICATION RELATED ***********/
/* This flag indicates new detection of increasing
   rate and pulsing attacks */
int INC = FALSE;

/*********** STATS RELATED ***********/
/* If this flag is set from command line 
   then packet statistics are written to a file */
int STATS = FALSE;
int gp[STAT_SIZE], gd[STAT_SIZE], bp[STAT_SIZE], bd[STAT_SIZE]; 
long attstart = -1;
int printed = 0;
long fw;
float INCstats, Dstats, Astats;
/* Files for gathering statistics on packets for charting */
FILE* stats = NULL;
FILE* rlstats = NULL;
FILE* timefile = NULL;

/* Prototype for packet monitoring function */
int run_monitor();

/* Function that prints packet statistics to a file */
void print_stats()
{
  int index = 0;
  int ct,i;
  while(TRUE)
    {
      if (STATS)
	{
	  /* Get one second statistics from kernel */
	  fw = getstats(FORWARDED);
	  fwp = getstats(FW_PACKETS);
	  fwb = getstats(FW_BYTES);

	  if (fw == NOT_FOUND)
	    {
	      if (REPLAY)
		timenow++;
	      break;
	    }
	  bp[index] = getstats(BAD_PASS);
	  bdtot = bd[index] = getstats(BAD_DROP);
	  gdtot = gd[index] = getstats(GOOD_DROP);
	  gp[index] = getstats(GOOD_PASS);
	  getstats(GOODPUT);
	  ct = time(0);
	}
      else
	{
	  /* Just to forward statistics */ 
	    getstats(FORWARDED);
	    getstats(BAD_PASS);
	    getstats(BAD_DROP);
	    getstats(GOOD_PASS);
	    getstats(GOOD_DROP);
	    getstats(GOODPUT);	 
	    getstats(FW_PACKETS);
	    getstats(FW_BYTES);
	    break;
	}
      index++;
    }
  for(i=0;i<index;i++)
    {
      char times[100];
      gettime(time(0), times);
      if (ABSOLUTE_TIME)
	fprintf(stats,"%ld %d %d %d %d %f\n",time(0)-index+i,
		gd[i], gp[i],
		bd[i], bp[i], Astats);
	  else
	    fprintf(stats,"%s %d %d %d %d %f\n",times,
		    gd[i], gp[i],
		    bd[i], bp[i], Astats);
      fflush(stats);
    }
}

/*********** DWARD MAIN ***********/
int dward_fd, sniff_fd;

/* Main program */
int main(int argc, char** argv)
{
  int i;
  pthread_t threadp,threaddef, threadpm, threadc, threadrt;
  FILE* prefixinfo;
  char timestamp[80];

  timenow = 0;
  bt = time(0);
  
  /* Initialize RNG */
  srand(time(0));
  
  /* Create timestamp for log files */
  strcpy(timestamp,ctime(&bt));
  /* Replace spaces with underscores,
     and colons with dashes */
  for(i=0;i<strlen(timestamp);i++)
    if(timestamp[i]==' ')
      timestamp[i] = '_';
    else if(timestamp[i]==':')
      timestamp[i] = '-';
  timestamp[strlen(timestamp)-1]=0;

  /* Take out options */
  for(i=0;i<argc;i++)
    {
      /* Most important messages */
      if (!strcmp(argv[i],"-d"))
	DEBUG = 1;
      /* More detailed messages that might clobber
	 processor under heavy load */
      else if (!strcmp(argv[i],"-dd"))
	DEBUG = 2;
      /* Statistics option */
      else if (!strcmp(argv[i],"-s"))
	{
	  STATS = TRUE;
	  strcpy(statsdir, argv[++i]);
	}
      else if (!strcmp(argv[i],"-p"))
	{
	  /* Prefix can also be specified from command line */
	  LOCAL_ADDRESS=inet_addr(argv[++i]);
	  LOCAL_MASK = atoi(argv[++i]);
	}
      /* Cossack option */
      else if (!strcmp(argv[i],"-C"))
	COSSACK = TRUE;
      else if (!strcmp(argv[i],"-D"))
	DefCOM = TRUE;
      /* Trace replay option */
      else if (!strcmp(argv[i],"-rep"))
	{
	  REPLAY = TRUE;
	  strcpy(tracefile_name,argv[++i]);
	}
      /* GUI option */
      else if(!strcmp(argv[i],"-g"))
	{
	  GUI = TRUE;
	  GUI_SERVER = FALSE;
	  gui_addr = inet_addr(argv[++i]);
	}
      else if(!strcmp(argv[i],"-gs"))
	{
	  GUI = TRUE;
	  GUI_SERVER = TRUE;
	}
      /* INC option */
      else if(!strcmp(argv[i],"-inc"))
	{
	  INC = TRUE;
	}
      /* Red team option */
      else if(!strcmp(argv[i],"-rt"))
	{
	  RED_TEAM = TRUE;
	  strcpy(statsdir, argv[++i]);
	}
      /* Absolute time option */
      else if(!strcmp(argv[i],"-abs"))
	{
	  ABSOLUTE_TIME = TRUE;
	}
      /* Monitor traffic option */
      else if(!strcmp(argv[i],"-a"))
	{
	  char a[20];
	  FILE* aggressive=fopen(argv[++i],"r");
	  if (!aggressive)
	    {
	      fprintf(stderr,"\tAGGRESSIVE option given but file"
		      " does not exist!\n");
	      exit(1);
	    }
	  while(TRUE)
	    {
	      fscanf(aggressive,"%s",a);
	      if (feof(aggressive))
		break;
	      aggressive_address[num_aggressive++]=inet_addr(a);
	    }
	}
    }
  /* Load prefix info */
  prefixinfo = fopen("configure/prefix.config","r");
  if (!prefixinfo)
    {
      fprintf(stderr,"\tNo prefiix.config file "
	      "in confugire directory!\n");	 
      exit(1);
    }
  /* Quick and dirty parsing */
  while(!feof(prefixinfo))
    {
      char a[255], left[80], right[80];
      int i=0, j=0, equal=FALSE;
      fgets(a,255,prefixinfo);
      while(i<strlen(a))
	{
	  if (a[i]=='=')
	    {
	      left[j]=0;
	      i++;
	      equal = TRUE;
	      j=0;
	    }
	  else if(!equal)
	    left[j++]=a[i++];
	  else if(isdigit(a[i]) || (a[i]=='.')) 
	    {
	      right[j++]=a[i++];
	    }
	  else
	    i++;
	}
      right[j]=0;
      trim(left);
      trim(right);
      if(strlen(left))
	{
	  if (!strcmp(left,"LOCAL_ADDRESS"))
	    {
	      if (LOCAL_ADDRESS == 0)
		LOCAL_ADDRESS=inet_addr(right);
	    }
	  if (!strcmp(left,"WATCHDOG_ADDRESS"))
	    WATCHDOG_ADDRESS=inet_addr(right);
	  else if(!strcmp(left,"LOCAL_MASK"))
	    {
	      if (LOCAL_MASK == -1)
		LOCAL_MASK = atoi(right);
	    }
	  else if(!strcmp(left,"MY_ADDRESS"))
	    strcpy(MY_ADDRESS,right);	  
	  else if(!strcmp(left,"GUI_PORT"))
	    GUIPORT = atoi(right);
	  else if(!strcmp(left,"WATCHDOG_PORT"))
	    WATCHDOG_PORT = atoi(right);
	}
    }
  /* Load dns info */  
  num_dns = 0;
  dnsfile = fopen("configure/dns_servers.txt","r");
  if (dnsfile)
    {
      while(dnsfile)
	{
	  char a[80];
	  fscanf(dnsfile,"%s",a);
	  fflush(stdout);
	  if (feof(dnsfile))
	  {
	    fclose(dnsfile);
	    break;
	  }
	  DNS_servers[num_dns++]=inet_addr(a);
	  printf("DNS %s %x\n",a,inet_addr(a));
	}
    }
  printf("Dns servers %d\n",num_dns);
  /* Create statistics directory */
  if (statsdir)
    {
      char path[100];
      sprintf(path,"%s/%s",statsdir,timestamp);
      mkdir(path,S_IRWXU);
    }
  /* Open statistics files */
  if (STATS)
    {
      char temp[80];
      sprintf(temp,"%s/%s/stats.txt",statsdir,timestamp);
      stats = fopen(temp,"w");
      sprintf(temp,"%s/%s/time.txt",statsdir,timestamp);
      timefile = fopen(temp,"w");
      sprintf(temp,"%s/%s/rlstats.txt",statsdir,timestamp);
      rlstats = fopen(temp,"w");
      sprintf(temp,"%s/%s/observed.txt",statsdir,timestamp);
      observed_stats = fopen(temp,"w");
      /* Fail gracefully if you have to */
      if (!stats || !timefile || !rlstats || !observed_stats)
	{
	  fprintf(stderr,"\tSTATS option given but files"
		  " cannot be created!\n");
	  fprintf(stderr,"\tCheck if stats directory exists "
		  "with write permission!\n");
	  exit(1);
	}
    }
  /* Open red team files */
  if (RED_TEAM)
    {
      char temp[80];
      sprintf(temp,"%s/%s/dward_coss_stats.txt",statsdir,timestamp);
      dward_coss_stats = fopen(temp,"w");
      sprintf(temp,"%s/%s/coss_stats.txt",statsdir,timestamp);
      coss_stats = fopen(temp,"w");
      sprintf(temp,"%s/%s/dward_stats.txt",statsdir,timestamp);
      dward_stats = fopen(temp,"w");
      sprintf(temp,"%s/%s/dward_stats_parsable.txt",statsdir,timestamp);
      dward_stats_parsable = fopen(temp,"w");
      /* Fail gracefully if you have to */
      if(!dward_coss_stats || !coss_stats 
	 || !dward_stats || !dward_stats_parsable)
	{
	  fprintf(stderr,"\tRED_TEAM option given but files "
		  "cannot be created!\n");
	  fprintf(stderr,"\tCheck if %s directory exists with "
		  "write permission!\n",statsdir);
	  exit(1);
	}
    }
  /* Open debug files */
  if (DEBUG || REPLAY)
    {
      flowdebug = fopen("debug/flow.txt","w");
      conndebug = fopen("debug/conn.txt","w");
      mondebug = fopen("debug/monitor.txt","w");
      classdebug = fopen("debug/class.txt","w");
      rldebug = fopen("debug/ratelimit.txt","w");
      if (GUI)
	guidebug = fopen("debug/guimessages.txt","w");
      /* Fail gracefully if you have to */
      if (!flowdebug || !conndebug || !mondebug 
	  || !classdebug || !rldebug || (GUI && !guidebug))
	{
	  fprintf(stderr,"\tDEBUG option given but files "
		  "cannot be created!\n");
	  fprintf(stderr,"\tCheck if debug directory exists with "
		  "write permission!\n");
	  exit(1);
	}
    }

  /* Open D-WARD device */  
  dward_fd = open("/dev/dward",O_RDONLY);
  if (dward_fd < 0) {
    fprintf(stderr,"Could not open dward device\n");
    exit(1);
  }
  
  /* Open sniff device */
  sniff_fd = open("/dev/sniff",O_RDONLY);
  if (sniff_fd < 0) {
    fprintf(stderr,"Could not open sniff device\n");
    exit(1);
  }
  
  /* Create all necessary hashes */
  initialize_flows();
  initialize_conns();
  initialize_rl();
  
  /* If DefCOM communication is enabled */  
  if (DefCOM)
    {
      struct sockaddr_in my_addr;    
      if ((defsock = socket(AF_INET, SOCK_DGRAM, UDP)) == -1) 
	{
	  perror("socket");
	  exit(1);
	}
      
      my_addr.sin_family = AF_INET;        
      my_addr.sin_port = htons(DEFPORT);     
      my_addr.sin_addr.s_addr = INADDR_ANY; 
      memset(&(my_addr.sin_zero), '\0', 8); 

      if (bind(defsock, (struct sockaddr *)&my_addr,
	       sizeof(struct sockaddr)) == -1) {
	perror("bind");
      }
    }

  /* If there is communication connect to multicast group */
  if (COSSACK)
    {
      int on,n;
      struct sockaddr_in my_addr;    
      if ((commfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) 
	{
	  perror("socket");
	  exit(1);
	}
      
      /* Set up my address */
      memset(&my_addr, 0, sizeof(struct sockaddr_in));
      my_addr.sin_family = AF_INET;          
      my_addr.sin_port = htons(WATCHDOG_PORT); 
      my_addr.sin_addr.s_addr = WATCHDOG_ADDRESS;
      
      setsockopt (commfd, SOL_SOCKET, SO_REUSEADDR, 
		  (void *) &on, sizeof (on));
      setsockopt (commfd, SOL_SOCKET, SO_REUSEPORT, 
		  (void *) &on, sizeof (on));	
      
      n=connect(commfd, (struct sockaddr *)&my_addr, sizeof my_addr);

      if (n<0)
	COSSACK = FALSE;

      if (RED_TEAM)
	{
	  fprintf(dward_coss_stats,"D-WARD connected to multicast group\n");
	  fflush(dward_coss_stats);
	}
    }
  
  /* Connect to GUI if needed */
  if (GUI)
    {
      struct sockaddr_in my_addr;    
      int on,n;
      char buf[255],temp[80];
      int numbytes;
      int i,j;
      if ((guifd = socket(AF_INET, SOCK_STREAM, 0)) == -1) 
	{
	  perror("socket");
	  exit(1);
	}
      
      /* Set up my address */
      memset(&my_addr, 0, sizeof(struct sockaddr_in));
      my_addr.sin_family = AF_INET;          
      my_addr.sin_port = htons(GUIPORT); 
      if (GUI_SERVER)
	my_addr.sin_addr.s_addr = INADDR_ANY; 
      else
	my_addr.sin_addr.s_addr = gui_addr; 
      setsockopt (guifd, SOL_SOCKET, SO_REUSEADDR, 
		  (void *) &on, sizeof (on));
      setsockopt (guifd, SOL_SOCKET, SO_REUSEPORT, 
		  (void *) &on, sizeof (on));	
       printf("Trying to connect to port %d ... ", GUIPORT);
       fflush(stdout);
      do
	{
	  n=connect(guifd, (struct sockaddr *)&my_addr, sizeof my_addr);
	}
      while (n<0);
      printf("Connected to GUI at %d\n",guifd);
      fflush(stdout);
      numbytes = recv(guifd,buf,1000,0);
      buf[numbytes] = 0;
      printf("Received victim addresses %s\n",buf);
      fflush(stdout);
      j=0;
      /* Take out victims' addresses */
      for(i=0;i<strlen(buf);i++)
	{
	  if (buf[i]==',')
	    {
	      temp[j]=0;
	      strcpy(victims[num_victims++],temp);
	      j=0;
	    }
	  else
	    temp[j++]=buf[i];
	}
    }  
  
  /* Start DefCOM thread */
  if (DefCOM)
    pthread_create(&threaddef, NULL, (void*)receive_defcom, NULL); 	
  
  /* Start communication thread */
  if (COSSACK)
    pthread_create(&threadc, NULL, (void*)multicast, NULL); 	
  
  /* Start processing thread */
  pthread_create(&threadp, NULL, (void*)process, NULL); 	
  
  /* Start replay */
  if (REPLAY)
    {
      tracefile=fopen(tracefile_name,"r");
      /* Fail gracefully if you have to */
      if (!tracefile)
	{
	  fprintf(stderr,"\tREPLAY option given but no tracefile %s"
		  " can be found!\n",tracefile_name);
	  exit(1);
	}
      pthread_create(&threadrt, NULL, (void*)read_trace, NULL); 
      pthread_join(threadrt,NULL);
    }

  /* Start monitoring thread */
  if (!REPLAY)
    {
      pthread_create(&threadpm, NULL, (void*)run_monitor, NULL); 	
      pthread_join(threadp,NULL);
    }
  return 0;
}






