/*
Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* System includes */ 
#include <sys/time.h>
#include <signal.h> 
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <math.h>
#include <linux/types.h>
#include <sys/errno.h> 
#include <string.h> 
#include <sys/wait.h> 
#include <pthread.h>

/* Our includes */
#include "replay.h"
#include "functions.h"
#include "constants.h"
#include "flow_stats.h"
#include "conn_stats.h"

/* Local address and mask */
extern __u32 LOCAL_ADDRESS;
extern int LOCAL_MASK;

/* Global time */
int starttime = 0;

/* Tracefile */
extern FILE* tracefile;

/* Function parse_time					       
   take out time from tcpdump output  			
   Input: tcpdump output in char array, int start position	
   Output: float time */
float parse_time(char* c, int* count)
{
	int i = *count, j;
	double val;
	char temp[10];
	for(i=0, j=0;i<strlen(c) && c[i]!=':';temp[j++]=c[i++]);
	if (i >= strlen(c))
	  return -1;
	i++;
	temp[j]=0;
	val = (double)atoi(temp);
	for(j=0;i<strlen(c) && c[i]!=':';temp[j++]=c[i++]);
	if (i >= strlen(c))
	  return -1;
	i++;
	temp[j]=0;
	val = val*60 + atoi(temp);
	for(j=0;i<strlen(c) && c[i]!='.';temp[j++]=c[i++]);
	if (i >= strlen(c))
	  return -1;
	i++;
	temp[j]=0;
	val = val*60 + atoi(temp);
	for(j=0;i<strlen(c) && !isspace(c[i]);temp[j++]=c[i++]);
	i++;
	temp[j]=0;
	val = val+(double)atoi(temp)/1000000;
	*count = i;
	return val;
}

/* Read data from tracefile in time order, parse it and
   update D-WARD statistics */
void read_trace()
{
  double oldtime=-1, nowtime, tracestarttime=-1, tracetime;
  struct timeval tv;
  char c[2000],active=FALSE;
  int count = 0;

  /* Wait for first read to finish */
  while(TRUE)
    {
      gettimeofday(&tv,0);
      nowtime = tv.tv_sec + (double)tv.tv_usec/1000000;
      if (oldtime == -1)
	oldtime = nowtime;
      count = 0;
      /* Get new line if the old one has been parsed */
      if (!active)
	{
	  if (fgets(c,2000,tracefile) == NULL && !feof(tracefile))
	    break;
	  if(feof(tracefile))
	    return;
	  c[strlen(c)-1]=0;
	}
      tracetime = parse_time(c, &count);
      if (tracestarttime == -1)
      tracestarttime = tracetime;
      if ((tracetime - tracestarttime) <= (nowtime - oldtime))
      {
	parse(c, count);
	active = FALSE;
      }
      else
      {
	active = TRUE;
	usleep(1);
	continue;
      }
    }
}

/* Parse IP address and port from tcpdump notation */
int parseIPAddress(char* c, unsigned int* address, unsigned short* port)
{
	int i, dots = 0, ja=0, jp=0;
	char tempa[17],tempp[6];
	for (i=0;i<strlen(c);i++)
	{
		if (c[i] == '.')
		{
		  dots++;
		  if (dots<4)
		    tempa[ja++] = c[i];
		}
		else 
		{
		  if(isdigit(c[i]))
		    {
		      if (dots<4)
			tempa[ja++] = c[i];
		      else
			tempp[jp++] = c[i];
		    }
		}
	}
	tempa[ja] = 0;
	tempp[jp] = 0;
	if (dots == 4 || dots == 3) 
	  {
	    if (strlen(tempa))
	      *address = inet_addr(tempa);
	    if (strlen(tempp))
	      *port = (short)atoi(tempp);
	    return 1;
	  }
	else
		return 0;
}

/* Parse tcpdump output, update cache  			
   Input: tcpdump output in char array, 
          starting index for parsing
   Output: void	*/
void parse(char* c, int counter)
{
	int protocol,srcmp,dstcmp, hdrlen, bytes;
	char t[80];
	unsigned int source, dest;
	unsigned short sport, dport;
	enum direction dir = NONE;
	enum pack_type ptype;
	__u32 mask = ((__u32)(~0)) >> (__u32)(32-LOCAL_MASK);
	eat_space(c,&counter);
	get_word(c,&counter,t);
	sport = 0;
	ptype = DP;

	/* Ignore ARP and Ethernet traffic */
	if (!parseIPAddress(t, &source, &sport))
	  {
	    printf("Not IP address %s\n",t);
	    return;
	  }
	eat_space(c,&counter);
	eat_word(c,&counter);
	eat_space(c,&counter);
	get_word(c,&counter,t);
	dport = 0;
	hdrlen = 20;
	bytes = 0;
	/* Ignore ARP and Ethernet traffic */
	if (!parseIPAddress(t,&dest, &dport))
	  {
	    printf("Not IP address %s\n",t);
	    return;
	  }
	eat_space(c,&counter);
	get_word(c,&counter,t);
	/* Account for bytes in TCP and UDP packets */
	if (!strcmp(t,"P") || !strcmp(t,"R") || !strcmp(t,"FP") 
	    || !strcmp(t,"S") || !strcmp(t,"F") || !strcmp(t,"."))
	{
	  protocol = IPPROTO_TCP;
	  hdrlen += 20;
	  /* Do nothing if it is ack, otherwise take byte info out */
	  eat_space(c,&counter);
	  get_word(c,&counter,t);
	  if(strcmp(t,"ack"))
	    {
	      char temp[30];
	      int start = FALSE, j=0,i;
	      ptype = ACK;

	      for(i=0;i<strlen(t);i++)
		{
		  if(t[i]=='(')
		    start = TRUE;
		  else if (isdigit(t[i]) && start)
		    temp[j++] = t[i];	  
		}
	      if (j)
		bytes += atoi(temp);
	    }
	}
	/* First parse ICMP data in case packets were truncated */
	else if (sport == 0 && dport == 0)
	{
	  protocol = IPPROTO_ICMP;
	  hdrlen += 4;
	}
	/* Assume it is UDP packet */
	else
	  {
	    char temp[30];
	    int start = FALSE, j=0,i;
	    hdrlen += 8;
	    protocol = IPPROTO_UDP;

	    /* First try parsing truncated UDP packets */
	    if (!strcmp(t,"udp"))
	      {
		eat_space(c,&counter);
		get_word(c,&counter,t);
		bytes += atoi(t);
	      }
	    else
	      {
		for(i=counter;i<strlen(c);i++)
		  {
		    if(c[i]=='(')
		      {
			start = TRUE;
			if (j)
			  bytes += atoi(temp);
			j=0;
		      }
		    else if(c[i]==')')
		      {
			start = FALSE;
			temp[j]=0;
		      }
		    else if (isdigit(c[i]) && start)
		      temp[j++] = c[i];	  
		  }
		if (j)
		  bytes += atoi(temp);
	      }
	    if(sport == NTP_PORT || dport == NTP_PORT)
	      {
		ptype = NTP;
	      }
	    if(sport == RP_PORT || dport == RP_PORT)
	      {
		ptype = STREAM;
	      }
	    if(sport == DNS_PORT || dport == DNS_PORT)
	      {
		if(sport == DNS_PORT)
		  {
		    ptype = DNSACK;
		  }
		else
		  {
		    ptype = DNS;
		  }
	      }
	  }

	srcmp = ((source & mask)
		 == (LOCAL_ADDRESS & mask));
	dstcmp = ((dest & mask) 
		  == (LOCAL_ADDRESS & mask)); 

	if (srcmp && !dstcmp)
	  {	
	    dir = OUT;
	  }
	else if (!srcmp && dstcmp)
	  {
	    dir = IN;
	  }

	if (dir == IN || dir == OUT)
	  {
	    if(srcmp)
	      {
		int class;
		class = update_conn_num(source,dest,sport,dport,dir,
				  hdrlen+bytes,protocol,ptype);
		update_flow_num(dest,dir, hdrlen+bytes, 
				protocol, 0, class, ptype);
	      }
	    else
	      {
		int class;
		class = update_conn_num(dest,source,dport,sport,dir,
					hdrlen+bytes,protocol,ptype);
		update_flow_num(source,dir, hdrlen+bytes, 
				protocol, 0, class, ptype);
	      }
	  }
}






