/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

import javax.swing.*;
import javax.swing.event.MouseInputAdapter;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;

/* Large panel that fills the main window */
public class DDoSPanel extends JPanel implements Runnable 
{
    private Dimension size; 
    int start_time;
    String events;

    private final Color colors[] = {
        Color.red, Color.blue, Color.green, Color.orange,
        Color.cyan, Color.magenta, Color.darkGray, Color.yellow};
		
    final int MAX_NODES = 5000;
    final int MAX_EDGES = 40000;
    static Font sans_serif6, sans_serif8;
    
    static {
	sans_serif6 = new Font("SansSerif", Font.PLAIN, 6);
	sans_serif8 = new Font("SansSerif", Font.PLAIN, 8);
    }

    int nnodes;
    Node nodes[] = new Node[MAX_NODES];
    int index=0;
    
    int nedges;
    Edge edges[] = new Edge[MAX_EDGES];

    int nsniffs;
    Sniff sniffs[] = new Sniff[MAX_EDGES];
    
    /* used for Zooming In/Out */
    private double scale = 1.0;
    boolean changed = false, right = false;
    boolean running = false;
    boolean active = false;
    boolean serverStarted = false;

    Thread relaxer;	
    
    Node pick=null;
    boolean pickfixed;
    boolean line = false;
    boolean move = false;
    Point where;
    Color BgColor; 
    SummaryFrame sumf;

    Image offscreen;      
    Dimension offscreensize;
    Graphics offgraphics;
    public static Animation animation;
    FontMetrics nodefm;

    int itemCount = 0;
    String message;
    int messageduration;
    int curtime;
    int playtime;

    Graph parent_graph;
    JPanel drawingArea;
    boolean paused;

    /* A message that is printed on the screen */
    public class MessageInstance{
	String message;
	/* When to print */
	int time;
	/* For how long */
	int duration;
	/* Command to be executed, currently does not work */
	String command;

	public MessageInstance(String s, int t, int d, String c)
	{
	    message = s;
	    time = t;
	    duration = d;
	    command = c;
	}
    }
    
    /* List of all messages */
    Vector MessageList;
    MessageInstance mi;
    
    /* Constructor */
    public DDoSPanel(Graph p) {
	parent_graph = p;
        setOpaque(true);
        size = new Dimension(0,0);
	curtime = 0;
	playtime = 0;
	message = "";
	messageduration = 0;
	MessageList = new Vector();
	BgColor = new Color(238,229,222);
	paused = false;
	events = new String();
	/* Set up the drawing area */
        drawingArea = new JPanel() {
		
		/* Function that paints/repaints the panel */
		protected void paintComponent(Graphics g) {
	
		    super.paintComponent(g);
		    Dimension d = getSize();
		    double display_bw, maxx=0, maxy=0;
		    Polygon domain;
		    
		    offscreensize = d;
		    offgraphics = g;
		    offgraphics.setFont(getFont());
		    
		    int fontsize = (int)(16+scale);
		    offgraphics.setFont(new Font
			("SansSerif", Font.PLAIN, fontsize));
		    offgraphics.setColor(getBackground());
		    offgraphics.fillRect(0, 0, d.width, d.height);
		    
		    /* Paint all edges first but only the visible part,
		       skip parts that are covered by nodes. This is so 
		       that nodes need not be redrawn every time */
		    for (int i = 0 ; i < nedges ; i++) {
			Edge e = edges[i];
			e.calculateIntersection(scale);
			offgraphics.setColor(e.edge_color);
			offgraphics.drawLine
			    (e.xint1, e.yint1, e.xint2, e.yint2);
		    }
		    /* Paint all sniffers */
		    for (int i = 0 ; i < nsniffs; i++) {
			Sniff s = sniffs[i];
			Edge e1 = s.fromedge;
			Edge e2 = s.toedge;
			int midx1 = (e1.xint1+e1.xint2)/2;
			int midy1 = (e1.yint1+e1.yint2)/2;
			int midx2 = (e2.xint1+e2.xint2)/2;
			int midy2 = (e2.yint1+e2.yint2)/2;
			Line2D.Double line1 = e1.getParallelLine(-10);
			Line2D.Double line2 = e2.getParallelLine(10);

			double x1 = line1.getX1();
			double x2 = line1.getX2();
			double y1 = line1.getY1();
			double y2 = line1.getY2();
			double x3 = line2.getX1();
			double x4 = line2.getX2();
			double y3 = line2.getY1();
			double y4 = line2.getY2();

			float ua = (float)
			    (((x4 - x3) * (y1 - y3) - 
			     (y4 - y3) * (x1 - x3))/
			    ((y4 - y3) * (x2 - x1) -
			     (x4 - x3) * (y2 - y1)));

			float ub = (float)
			    (((x2 - x1) * (y1 - y3) - 
			     (y2 - y1) * (x1 - x3))/
			    ((y4 - y3) * (x2 - x1) -
			     (x4 - x3) * (y2 - y1)));


			/* Use Kramers rule to compute xi and yi */
			int xi = (int)(x1 + ua * (x2 - x1));
			int yi = (int)(y1 + ub * (y2 - y1));

			System.out.println("Intersection between "+
					   line1.getX1()+ " "+line1.getY1()+
					   " to "+
					   line1.getX2()+ " "+line1.getY2()+
					   " and " +
					   line2.getX1()+ " "+line2.getY1()+
					   " to "+
					   line2.getX2()+ " "+line2.getY2()+
					   " is "+
					   xi + " " + yi);

			offgraphics.setColor(s.sniff_color);
			offgraphics.drawLine
			    (midx1,midy1,xi,yi);
			offgraphics.drawLine
			    (xi,yi,midx2, midy2);
		    }
		    nodefm = offgraphics.getFontMetrics();
		    
		    /* Repaint all nodes */
		    for (int i = 0 ; i < nnodes ; i++) 
			{
			    nodes[i].paintNode
				(offgraphics, nodes[i], nodefm, 
				 scale, scale, 0, 0, true, playtime);
			}
		}
	    };
        drawingArea.setBackground(BgColor);
        drawingArea.addMouseListener(new MyMouseListener(this));
	drawingArea.addMouseMotionListener(new MyMouseListener(this));
        
	/* Put the drawing area in a scroll pane */
        JScrollPane scroller = new JScrollPane(drawingArea);
        scroller.setPreferredSize(new Dimension(200,200));

        /* Layout this demo */
        setLayout(new BorderLayout());
        add(scroller, BorderLayout.CENTER);
    }

    /* Mouse listener captures and responds to mouse events */
    class MyMouseListener extends MouseInputAdapter {
 
	DDoSPanel ddos_panel;

	public MyMouseListener(DDoSPanel panel)
	{
	    ddos_panel = panel;
	}

	public void mousePressed(MouseEvent evt) {
	    if (SwingUtilities.isRightMouseButton(evt) == true)
		right = true;
	    move = false;
	    line = false;
	    int x = evt.getX();
	    int y = evt.getY();
	    int scaled_x, scaled_y;
	    boolean changed = false;
	    long address;
	    
	    scaled_x = (new Double(x * (1/scale))).intValue();
	    scaled_y = (new Double(y * (1/scale))).intValue();
	    	    
	    /* This for loop tries to find the target node 
	       if mouse is pressed within the node */
	    for (int i = 0 ; i < nnodes ; i++) {
		Node n = nodes[i];
		if ((n.x-n.width/2) <= scaled_x 
		    && (n.x+n.width/2) >= scaled_x && 
		    (n.y-n.height/2) <= scaled_y 
		    && (n.y+n.height/2) >= scaled_y)
		    {
			pick = n;
			index=i;
		    }
	    }
	    if (sumf != null)
		sumf.show();
	    if (pick == null || right == true)
	    {
		/* Redraw main window */
		if (parent_graph.sf != null)
		    parent_graph.sf.show();
		
		/* Refresh all TrackPanels */
		if (pick == null)
		    {
			for (int i = 0 ; i < nnodes ; i++) 
			    {
				Node n = nodes[i];
				if (n.panel != null && n.panel.active == true)
				    {
					n.panel.isVisible = true;
					n.tf.show();
				    }
			    }
			return;
		    }
		/* If right button is pressed within the node open 
		   its TrackPanel */
		if (right)
		    {
			if (pick != null)
			    {
				/* If window is open don't open it again */
				if (pick.isTracking == false)
				    {
					pick.isTracking = true;
					pick.tf.addWindowListener(pick.tf);
					pick.tf.addComponentListener(pick.tf);
					pick.tf.pack(); 		
					pick.tf.setSize(300,300);
					pick.tf.setLocation
					    (new Point((int)pick.x,
						       (int)pick.y));
				    }
				pick.panel.active = true;
				pick.tf.show();             
				pick.panel.isVisible = true;
			    }
			pick = null;	
			changed = true;
			right = false;
		    }
	    }
	} /* mousePressed() */ 

	/* Move the selected node */
	public void mouseDragged(MouseEvent evt) {
	    int x = evt.getX();	 
	    int y = evt.getY();
	    
	    int scaled_x, scaled_y;
	    
	    if (right)
		return;
	    scaled_x = (new Double(x * (1/scale))).intValue();
	    scaled_y = (new Double(y * (1/scale))).intValue();
	    
	    if (pick == null)
		return;
	    
	    if (line) {
		where.x = scaled_x;
		where.y = scaled_y;
	    }
	    else if (!right){
		pick.x = scaled_x;
		pick.y = scaled_y;
		boolean changed = false;
		if ((int)pick.x + 100 > ddos_panel.size.width)
		    {
			ddos_panel.size.width = (int)pick.x + 100;
		    }
		if ((int)pick.y + 100 > ddos_panel.size.height)
		    {
			ddos_panel.size.height = (int)pick.y + 100;
		    }
		ddos_panel.update();
	    }
	} /* mouseDragged () */
	
	/* This needs to be defined */
	public void mouseClicked(MouseEvent evt) {
	} /* mouseClicked () */
	
	
	/* If we were dragging the component, 
	   place it at new position */
	public void mouseReleased(MouseEvent evt) 
	{
	    int x = evt.getX();
	    int y = evt.getY();
	    
	    int scaled_x, scaled_y;
	    /*
	      scaled_x = (new Double(x * (1/scale))).intValue();
	      scaled_y = (new Double(y * (1/scale))).intValue();
	      
	      if (pick == null)
	      return;
	      Node newPick=pick;
	      double bestdist = Double.MAX_VALUE;
	      for (int i = 0 ; i < nnodes ; i++) 
	      {
	      Node n = nodes[i];
	      double dist = (n.x - scaled_x) * (n.x - scaled_x) +
	      (n.y - scaled_y) * (n.y - scaled_y);
	      if (dist < bestdist) {
	      newPick = n;
	      bestdist = dist;
	      }
	      }
	      
	      pick.fixed = pickfixed;
	      if (right == false)
	      {
	      pick = null;
	      }
	    */
	    if (right)
		{
		    changed = true;
		    right = false;
		}
	    line = false;
	    
	} /* mouseReleased() */	
    }
    
    /* When start button is pressd animation starts */ 
    public void startServer(String addresses)
    {
	System.out.println("Starting server");
	if (!serverStarted)
	    {
		animation = new Animation(parent_graph, this, 7777, addresses);
		serverStarted = true;
	    }
    }
    
    /* Update the panel by redrawing the area */
    public void update()
    {
	drawingArea.setPreferredSize(size);
	
	/* Let the scroll pane know to update itself
	   and its scrollbars */
	drawingArea.revalidate();
	drawingArea.repaint();
    }
    
    /* Add new message to the list */
    public void InsertMessage(String s, int t, int d, String c)
    {
	MessageInstance m = new MessageInstance(s,t,d,c);
	MessageList.addElement(m);
    }
    
    /* Activate message whose time has come */
    public void ReleaseMessage()
    {
	for(int i=0;i<MessageList.size();i++)
	    {
		MessageInstance m = (MessageInstance)(MessageList.get(i));
		if (m.time == playtime-2)
		    {
			this.mi = m;
			int time = playtime - 2;
			int sec = time%60;
			int min = (time/60)%60;
			int hour = time/3600;
			String tm =  new String();			
			if (hour != 0)
				tm += hour+":";
			if (min != 0)
				tm += +min+":";
			tm += sec+" ";
			events += (tm + m.message + "\n");

		    }
	    }
    }
        
    /* Draw selected message on the screen */
    public void DrawMessage()
    {
	if (paused)
	    return;
	Graphics g = drawingArea.getGraphics();
	Dimension d = getSize();
	Font fnt = new Font("Arial",Font.PLAIN,16);
	g.setFont(fnt);
	FontMetrics fm = g.getFontMetrics();
	int h = fm.getHeight()+60;
	if (mi!=null || messageduration == 0)
	    {
		/* Time to change the message */
		if (!message.equals(""))
		    {
			g.setColor(BgColor);
			g.drawString
			    (message,(d.width-fm.stringWidth(message))/2,h);
		    }
		if (mi!=null)
		    {
			message = mi.message;
			messageduration = mi.duration;
			mi = null;
		    }
		else
		    message = "";
	    }
	if (messageduration>0)
	    {
		g.setColor(Color.blue);
		g.drawString(message,(d.width-fm.stringWidth(message))/2,h);
		messageduration--;
	    }
    }

    /* Paint packets on the edges */
    public void paintEdges()
    {
	if (!active)
	    return;
	Graphics g = drawingArea.getGraphics();

	/* We allow one second delay for animation */
	int ct = playtime - 1;
	int pt = playtime - 2;
	
	/* For all edges */
	for (int i = 0 ; i < nedges ; i++) 
	    {
		Edge e = edges[i];
		
		if (pt<0 || ct<0)
		    return;
		
		if (((e.packets.size() <= pt ) 
		     || (e.packets.get(pt)==null)) 
		    && ((e.packets.size() <= ct ) 
			|| (e.packets.get(ct)==null))
		    && ((e.replies.size() <= pt ) 
			|| (e.replies.get(pt)==null))
		    && ((e.replies.size() <= ct ) 
			|| (e.replies.get(ct)==null)))							
		    continue;
		
		
		int x10,x11,x12,x13,y10,y11,y12,y13;
		int d1 = 5;
		int d2 = 5;
		
		Line2D.Double line = e.getParallelLine(-10);
		Line2D.Double liner = e.getParallelLine(10);

		Line2D.Double sentline = e.getParallelLine(-40);
		Line2D.Double recline = e.getParallelLine(40);
		Line2D.Double sentlineB = e.getParallelLine(-25);
		Line2D.Double reclineB = e.getParallelLine(55);
		
		Font fnt = new Font("Arial",Font.BOLD,10);
		g.setFont(fnt);
		
		/* Erase packets to the victim */
		if (e.packets.get(pt)!=null)
		    {
			int packets 
			    = ((Integer)e.packets.get(pt)).intValue();
			int bytes = ((Integer)e.bytes.get(pt)).intValue();
			int apackets = 0;
			int abytes = 0;
			if ((e.apackets.size() > pt) && (e.apackets.get(pt) != null))
			{
			    apackets = ((Integer)e.apackets.get(pt)).intValue();
			    abytes = ((Integer)e.abytes.get(pt)).intValue();
			}
			if (packets>0)
			    {
				int xint1 = (int)line.getX1();
				int yint1 = (int)line.getY1();
				int xint2 = (int)line.getX2();
				int yint2 = (int)line.getY2();
				
				int len = 
				    (int)Math.abs(Math.sqrt((xint1-xint2)
							    *(xint1-xint2)
							    +(yint1-yint2)
							    *(yint1-yint2)));
				int numo = packets+1; 
				if (numo> len/10)
				    numo = len/10;		    
				if (numo <= 1)
				    numo = 2;
				double midpoint_xo 
				    = (double)(xint2-xint1)/numo;
				double midpoint_yo 
				    = (double)(yint2-yint1)/numo;
				double mid_x
				    = (double)(sentline.getX2()+sentline.getX1())/2;
				double mid_y 
				    = (double)(sentline.getY2()+sentline.getY1())/2;
				double mid_xB 
				    = (double)(sentlineB.getX2()+sentlineB.getX1())/2;
				double mid_yB 
				    = (double)(sentlineB.getY2()+sentlineB.getY1())/2;				
				
				g.setColor(BgColor);
				String sp = new String();
				String sb = new String();
				if ((packets - apackets) > 0 && (bytes-abytes) > 0)
				    {
					sp = new String(""+(packets-apackets));
					sb = new String(""+(bytes-abytes)+" B");
					g.drawString
					    (sp,(int)mid_x,(int)mid_y);
					g.drawString
					    (sb,(int)mid_x,(int)mid_yB);
				    }
				FontMetrics fm = offgraphics.getFontMetrics();
				if (apackets > 0)
				    {
					g.drawString
					    (""+apackets,(int)mid_x+fm.stringWidth(sp),(int)mid_y);
					g.drawString
					    (""+abytes+" B",(int)mid_x+fm.stringWidth(sb),(int)mid_yB);
				    }
				for (int pos=1;pos<numo;pos++)
				    {
					x10=(int)(xint1+pos*midpoint_xo
						  +e.offset
						  *Math.cos(e.alpha));
					y10=(int)(yint1+pos*midpoint_yo
						  +e.offset
						  *Math.sin(e.alpha));
					x11=x10+(int)Math.floor
					    (d1*scale*Math.cos(e.alpha));
					y11=y10-(int)Math.floor
					    (d1*scale*Math.sin(e.alpha));
					x12=x11-(int)Math.floor
					    (d2*scale*Math.sin(e.alpha));
					y12=y11-(int)Math.floor
					    (d2*scale*Math.cos(e.alpha));
					x13=x10-(int)Math.floor
					    (d2*scale*Math.sin(e.alpha));
					y13=y10-(int)Math.floor
					    (d2*scale*Math.cos(e.alpha));
					int xarray[]={x10, x11, x12, x13};
					int yarray[]={y10, y11, y12, y13};
					g.drawPolygon(xarray, yarray, 4);
					g.fillPolygon(xarray, yarray, 4);
				    }
			    }
		    }
		/* Erase replies from the victim */ 
		if (e.replies.get(pt)!=null)
		    {
			int packets 
			    = ((Integer)e.replies.get(pt)).intValue();
			int bytes 
			    = ((Integer)e.rbytes.get(pt)).intValue();
			if (packets>0)
			    {
				int xint1 = (int)liner.getX1();
				int yint1 = (int)liner.getY1();
				int xint2 = (int)liner.getX2();
				int yint2 = (int)liner.getY2();
				
				int len 
				    = (int)Math.abs(Math.sqrt
						    ((xint1-xint2)
						     *(xint1-xint2)
						     +(yint1-yint2)
						     *(yint1-yint2)));
				int numo = packets+1;
				if (numo> len/10)
				    numo = len/10;		    
				if (numo <= 1)
				    numo = 2;
				double midpoint_xo 
				    = (double)(xint2-xint1)/numo;
				double midpoint_yo 
				    = (double)(yint2-yint1)/numo;
				double mid_x
				    = (double)(recline.getX2()+recline.getX1())/2;
				double mid_y 
				    = (double)(recline.getY2()+recline.getY1())/2;
				double mid_xB 
				    = (double)(reclineB.getX2()+reclineB.getX1())/2;
				double mid_yB 
				    = (double)(reclineB.getY2()+reclineB.getY1())/2;

				g.setColor(BgColor);
				Node wn;
				wn = e.tnode;
				g.drawString
				    (""+packets,(int)mid_x,(int)mid_y);
				g.drawString
				    (""+bytes+" B",(int)mid_x, (int)mid_yB);
				for (int pos=1;pos<numo;pos++)
				    {
					x10=(int)(xint1+pos*midpoint_xo
						  +e.offset*Math.cos(e.alpha));
					y10=(int)(yint1+pos*midpoint_yo
						  +e.offset*Math.sin(e.alpha));
					x11=x10+(int)Math.floor
					    (d1*scale*Math.cos(e.alpha));
					y11=y10-(int)Math.floor
					    (d1*scale*Math.sin(e.alpha));
					x12=x11-(int)Math.floor
					    (d2*scale*Math.sin(e.alpha));
					y12=y11-(int)Math.floor
					    (d2*scale*Math.cos(e.alpha));
					x13=x10-(int)Math.floor
					    (d2*scale*Math.sin(e.alpha));
					y13=y10-(int)Math.floor
					    (d2*scale*Math.cos(e.alpha));
					int xarray[]={x10, x11, x12, x13};
					int yarray[]={y10, y11, y12, y13};
					g.drawPolygon(xarray, yarray, 4);
					g.fillPolygon(xarray, yarray, 4);
				    }
			    }
		    }
		g.setColor(e.edge_color);
		g.drawLine(e.xint1, e.yint1, e.xint2, e.yint2);
		
		/* Draw packets to the victim */
		if ((e.packets.size() > ct) && (e.packets.get(ct) != null))
		    {
			int packets 
			    = ((Integer)e.packets.get(ct)).intValue();
			int bytes 
			    = ((Integer)e.bytes.get(ct)).intValue();
			int apackets = 0;
			int abytes = 0;
			if ((e.apackets.size() > ct) && (e.apackets.get(ct) != null))
			{
			    apackets = ((Integer)e.apackets.get(ct)).intValue();
			    abytes = ((Integer)e.abytes.get(ct)).intValue();
			}
 			if (packets>0)
			    {
				int xint1 = (int)line.getX1();
				int yint1 = (int)line.getY1();
				int xint2 = (int)line.getX2();
				int yint2 = (int)line.getY2();
				
				int len = (int)Math.abs(Math.sqrt
							((xint1-xint2)
							 *(xint1-xint2)
							 +(yint1-yint2)
							 *(yint1-yint2)));
				int num = packets+1;
				if (num> len/10)
				    num = len/10;
				if (num <= 1)
				    num = 2;
				int anum = 0;
				if (apackets > 0)
				  anum = (int)((double)num/packets*apackets);
				double midpoint_x 
				    = (double)(xint2-xint1)/num;
				double midpoint_y 
				    = (double)(yint2-yint1)/num;
				double mid_x 
				    = (double)(sentline.getX2()+sentline.getX1())/2;
				double mid_y 
				    = (double)(sentline.getY2()+sentline.getY1())/2;
				double mid_xB 
				    = (double)(sentlineB.getX2()+sentlineB.getX1())/2;
				double mid_yB 
				    = (double)(sentlineB.getY2()+sentlineB.getY1())/2;
				
				g.setColor(new Color(70,130,196));
				String sp = new String();
				String sb = new String();
				if ((packets - apackets) > 0 && (bytes-abytes) > 0)
				    {
					sp = new String(""+(packets-apackets));
					sb = new String(""+(bytes-abytes)+" B");
					g.drawString
					    (sp,(int)mid_x,(int)mid_y);
					g.drawString
					    (sb,(int)mid_x,(int)mid_yB);
				    }
				g.setColor(new Color(200,0,0));
				FontMetrics fm = offgraphics.getFontMetrics();
				if (apackets > 0)
				    {
					g.drawString
					    (""+apackets,(int)mid_x+fm.stringWidth(sp),(int)mid_y);
					g.drawString
					    (""+abytes+" B",(int)mid_x+fm.stringWidth(sb),(int)mid_yB);
				    }
				for (int pos=1;pos<num;pos++)
				    {
					x10=(int)(xint1+pos*midpoint_x
						  -e.offset*Math.cos(e.alpha));
					y10=(int)(yint1+pos*midpoint_y
						  -e.offset*Math.sin(e.alpha));
					x11=x10+(int)Math.floor
					    (d1*scale*Math.cos(e.alpha));
					y11=y10-(int)Math.floor
					    (d1*scale*Math.sin(e.alpha));
					x12=x11-(int)Math.floor
					    (d2*scale*Math.sin(e.alpha));
					y12=y11-(int)Math.floor
					    (d2*scale*Math.cos(e.alpha));
					x13=x10-(int)Math.floor
					    (d2*scale*Math.sin(e.alpha));
					y13=y10-(int)Math.floor
					    (d2*scale*Math.cos(e.alpha));
					int xarray[]={x10, x11, x12, x13};
					int yarray[]={y10, y11, y12, y13};
					if (anum > 0)
					  {
					    g.setColor(new Color(200,0,0));
					    anum --;
					  }
					else
					    g.setColor(new Color(70,130,196));
					g.drawPolygon(xarray, yarray, 4);
					g.fillPolygon(xarray, yarray, 4);
				    } 
			    } 
		    }
		
		/* Draw replies from the victim */
		if ((e.replies.size() > ct) && (e.replies.get(ct) != null))
		    {
			int packets 
			    = ((Integer)e.replies.get(ct)).intValue();
			int bytes 
			    = ((Integer)e.rbytes.get(ct)).intValue();
			if (packets>0)
			    {
				int xint1 = (int)liner.getX1();
				int yint1 = (int)liner.getY1();
				int xint2 = (int)liner.getX2();
				int yint2 = (int)liner.getY2();
				
				int len = (int)Math.abs(Math.sqrt
							((xint1-xint2)
							 *(xint1-xint2)
							 +(yint1-yint2)
							 *(yint1-yint2)));
				int num = packets+1;
				if (num>len/10)
				    num = len/10;
				if (num <= 1)
				    num = 2;
				double midpoint_x 
				    = (double)(xint2-xint1)/num;
				double midpoint_y 
				    = (double)(yint2-yint1)/num;
				double mid_x
				    = (double)(recline.getX2()+recline.getX1())/2;
				double mid_y 
				    = (double)(recline.getY2()+recline.getY1())/2;
				double mid_xB 
				    = (double)(reclineB.getX2()+reclineB.getX1())/2;
				double mid_yB 
				    = (double)(reclineB.getY2()+reclineB.getY1())/2;

				g.setColor(new Color(102,204,51));
				g.drawString
				    (""+packets,(int)mid_x,(int)mid_y);
				g.drawString
				    (""+bytes+" B",(int)mid_x,(int)mid_yB);
				
				for (int pos=1;pos<num;pos++)
				    {
					x10=(int)(xint1+pos*midpoint_x
						  -e.offset*Math.cos(e.alpha));
					y10=(int)(yint1+pos*midpoint_y
						  -e.offset*Math.sin(e.alpha));
					x11=x10+(int)Math.floor
					    (d1*scale*Math.cos(e.alpha));
					y11=y10-(int)Math.floor
					    (d1*scale*Math.sin(e.alpha));
					x12=x11-(int)Math.floor
					    (d2*scale*Math.sin(e.alpha));
					y12=y11-(int)Math.floor
					    (d2*scale*Math.cos(e.alpha));
					x13=x10-(int)Math.floor
					    (d2*scale*Math.sin(e.alpha));
					y13=y10-(int)Math.floor
					    (d2*scale*Math.cos(e.alpha));
					int xarray[]={x10, x11, x12, x13};
					int yarray[]={y10, y11, y12, y13};
					g.drawPolygon(xarray, yarray, 4);
					g.fillPolygon(xarray, yarray, 4);
				    }  
			    }
		    }
		
		/* Changing the offset gives the impression of packets 
		   flowing along the edges */  
		if (!paused)
		    {
			if (e.offset==-3)
			    e.offset = 3;
			else
			    e.offset = -3;
		    }
	    }	
	for (int i = 0 ; i < nnodes ; i++) 
	    DrawDrops(nodes[i]);
	/* See if some node has received multicast message */
	for (int i=0;i<nnodes;i++)
	    if (nodes[i].multicast == true)
		{
		    Font fnt = new Font("SansSerif",Font.PLAIN,18);
		    g.setFont(fnt);
		    nodes[i].paintNode
			(g, nodes[i], nodefm, 
			 scale, scale, 0, 0, true, ct);
		}
	printSummary();
    } /* paintEdges */
    
    /* Thread run, defines actions at every second */
    public void run() {
	while (running) {
	    //System.out.println("Running is "+running+" time is "+curtime);
	    if (animation == null)
		continue;
	    animation.ReadTimeUpdate();
	    try{
		if(start_time!=-1 && active)
		    {
			curtime++;
			/* Paint the current time on the panel */
			if (!paused)
			    {
				playtime++;
				paintTime(playtime-1, playtime-2);
				paintEdges();
			    }
		    }
		ReleaseMessage();
		DrawMessage();
		Thread.sleep(1000);
	    }
	    catch(Exception e)
		{
		    System.out.println("Exception "+e.toString());
		}
	}
    } /* run */
    
    /* Start the thread */
    public void start() {
	start_time = (int)(System.currentTimeMillis()/1000);
	for(int i=0;i<nnodes;i++)
	    nodes[i].panel.start();
	running = true;
	active = true;
	if (!serverStarted)
	    {
		System.out.println("Start thread");
		relaxer = new Thread(this);
		relaxer.start();
		sumf = new SummaryFrame();
		sumf.pack();
		sumf.setLocation(new Point(570,470));
		sumf.show();
	    }
    }
    
    public void pause()
    {
	if (paused)
	    {
		paused = false;
		for(int i=0;i<nnodes;i++)
		    nodes[i].panel.paused = false;
	    }
	else
	    {
		paused = true;
		for(int i=0;i<nnodes;i++)
		    nodes[i].panel.paused = true;
	    }
    }

    /* Handle message that delives info about flow on a certain edge */
    void handleFlowMessage
	(int time, String source, String dest, int sentp, 
	 int recp, int sent, int rec, int attp, int att)
    {
	if (active)
	    {
		int index = findNode(source);
		int dindex = findNode(dest);
		
		/* Find the edge in question */
		int edgeindex = findEdgeTo(source,dest);
		if (edgeindex != -1)
		    {
			/* shoud be time-start-time for synchronized clocks */
			edges[edgeindex].InsertPacket(sentp, curtime, sent);
			edges[edgeindex].InsertReply(recp, curtime, rec); 
			edges[edgeindex].InsertAttackPacket(attp, curtime, att);  
		    }
		else
		    System.out.println("Did not find edge between "+source+" and "+dest);
		if (index == -1 || dindex==-1)
		    return;
		else
		    {
			int iindex = nodes[index].findNeighbor(nodes[dindex]);
			/* Update number above the node */
			nodes[index].total_goodpass += sent;
			nodes[index].total_badpass += att;
			nodes[index].panel.handleFlowMessage
			    ((curtime), source, iindex, sent, rec);  
			/* shoud be time-start-time for synchronized clocks */
		    }
	    }
	return;
    } /* handleFlowMessage */
    
    /* Just add data to associated node, do not update the edge */
    void handleFlowMessage
	(int time, String source, int Ninterface, int sent, int rec)
    {
	if (active)
	    {
		int index = findNode(source);
		if (index == -1)
		    return;
		else
		    nodes[index].panel.handleFlowMessage
			((curtime), source, Ninterface,sent,rec);
	    }
    }
    
    /* Handle message about rate limits */
    void handleRateMessage
	(int time, String source, String node, double ratelim, 
	 int sentp, int recp, int sent, int rec, int attp, int att, 
	 int gooddrop, int baddrop)
    {
	if (active)
	    {
		int index = findNode(source);
		int vindex = findNode(node);

		if (ratelim>0)
		    {
			if (nodes[index].limited == false)
			    {
				InsertMessage(new String(nodes[index].label+" "+nodes[index].ID+" rate limited flow to "+nodes[vindex].type+" "+nodes[vindex].ID),curtime-1,5,"");
				nodes[index].limited = true;
			    }
		    }
		else if (nodes[index].limited == true)
		    {
			InsertMessage(new String(nodes[index].label+" "+nodes[index].ID+" removing rate limit on "+nodes[vindex].type+" "+nodes[vindex].ID),curtime+1,5,"");
			nodes[index].limited = false;
		    }
		
		int edgeindex = findEdgeTo(source,node);
		if (edgeindex != -1)
		    {
			/* shoud be time-start-time for synchronized clocks */
			edges[edgeindex].InsertPacket(sentp, curtime, sent);
			edges[edgeindex].InsertReply(recp, curtime, rec);  
			edges[edgeindex].InsertAttackPacket(attp, curtime, att);   
		    }
		if (index == -1)
		    return;
		else
		    nodes[index].panel.handleRateMessage
			(curtime, node, ratelim, sent, rec);
		if (curtime>2)
		    nodes[index].insertDrop(curtime-2,gooddrop,baddrop);
		if (sent > 0)
		    nodes[index].total_goodpass += sent;
		if (att > 0)
		    nodes[index].total_badpass += att;
	    }
	return;
    }
 
    /* Convert a number to KB OR MB */
    String toBytes(int num)
    {
	String temp;
	if (num < 1000)
	    temp = new String(""+num+" B");
	else if (num < 1000000)
	    temp = new String(""+Math.floor((double)num/100)/10+" KB");
	else
	    temp = new String(""+Math.floor((double)num/100000)/10+" MB");
	return temp;
    }

    /* Print summary of passed and dropped packets */
    void printSummary()
    {
	if (paused)
	    return;
	String summary = new String();
	for(int i=0;i<nnodes;i++)
	    if (nodes[i].type.startsWith("router"))
		{
		    double gp = Math.floor((double)nodes[i].total_goodpass/(nodes[i].total_goodpass+nodes[i].total_gooddrop)*10000)/100;
		    double gd = Math.floor( (double)nodes[i].total_gooddrop/(nodes[i].total_goodpass+nodes[i].total_gooddrop)*10000)/100;
		    double bp = Math.floor((double)nodes[i].total_badpass/(nodes[i].total_badpass+nodes[i].total_baddrop)*10000)/100;
		    double bd = Math.floor((double)nodes[i].total_baddrop/(nodes[i].total_badpass+nodes[i].total_baddrop)*10000)/100;
		    String good = null, bad = null;
		    if (nodes[i].total_goodpass+nodes[i].total_gooddrop > 0)
			good = new String("Good traffic \n\t forwarded "+toBytes(nodes[i].total_goodpass) +" ("+gp
					   +"%) "+"\n\t dropped  "+toBytes(nodes[i].total_gooddrop)+" ("+gd+"%) \n");
		    if(nodes[i].total_badpass+nodes[i].total_baddrop > 0)
			bad = new String("Attack traffic \n\t forwarded "+toBytes(nodes[i].total_badpass) + " ("+bp
					    +"%) "+"\n\t dropped  "+toBytes(nodes[i].total_baddrop)+" ("+bd+"%) \n");
		    if (good != null || bad != null)
			summary += new String(nodes[i].label + " " +nodes[i].ID+":\n");
		    if (good != null) 
			summary += good;
		    if (bad != null)
			summary += bad;
		}
	summary += events;
	sumf.DrawMessage(summary);
    }

    /* Paint the node which received multicast message */
    void receivedMulticast(String source)
    {
	if (active)
	    {
		int index = findNode(source);
		nodes[index].receivedMulticast(drawingArea.getGraphics(), curtime);
	    }
    }

    /* Paint the node which sent multicast message */
    void sentMulticast(String source)
    {
	if (active)
	    {
		int index = findNode(source);
		nodes[index].sentMulticast(drawingArea.getGraphics(), curtime);
	    }
    }

    /* Draw dropped bytes */
    void DrawDrops(Node n)
    {
	int x = (int)n.x;
	int y = (int)n.y + 100;
	Graphics g = drawingArea.getGraphics();
	Font fnt = new Font("Arial",Font.BOLD,10);
	g.setFont(fnt);

	int ct = playtime - 3;
	int pt = playtime - 4;
	System.out.println("In draw drops ct "+ct+" pt "+pt+" sizes "+n.gooddrops.size()+" "+n.baddrops.size());
	if (ct < 0 || pt < 0)
	    return;
	/* Erase old drops */
	if (pt >= 0)
	    {
		g.setColor(BgColor);
		if ((n.gooddrops.size() > pt) && (n.gooddrops.get(pt) != null))
		    {
			int gooddrop = ((Integer)n.gooddrops.get(pt)).intValue();
			if (gooddrop > 0)
			    {
				g.fillRect(x,y,(int)Math.log(gooddrop)*10,5);
				g.drawString(""+gooddrop,x+(int)Math.log(gooddrop)*10,y);
			    }
		    }
		if ((n.baddrops.size() > pt) && (n.baddrops.get(pt) != null))
		    {
			int baddrop = ((Integer)n.baddrops.get(pt)).intValue();
			if (baddrop > 0)
			    {
				g.fillRect(x,y+10,(int)Math.log(baddrop)*10,5);
				g.drawString(""+baddrop,x+(int)Math.log(baddrop)*10,y+10);
			    }
		    }
	    }
	if ((n.gooddrops.size() > ct) && (n.gooddrops.get(ct) != null))
	    {
		int gooddrop = ((Integer)n.gooddrops.get(ct)).intValue();
		if (gooddrop > 0)
		    {
			g.setColor(new Color(70,130,196));
			g.fillRect(x,y,(int)Math.log(gooddrop)*10,5);
			g.drawString(""+gooddrop,x+(int)Math.log(gooddrop)*10,y);
		    }
	    }
	if ((n.baddrops.size() > ct) && (n.baddrops.get(ct) != null))
	    {
		int baddrop = ((Integer)n.baddrops.get(ct)).intValue();
		if (baddrop > 0)
		    {
			g.setColor(Color.red);
			g.fillRect(x,y+10,(int)Math.log(baddrop)*10,5);
			g.drawString(""+baddrop,x+(int)Math.log(baddrop)*10,y+10);
		    }
	    }
    }
    
    /* Find edge between two nodes */
    int findEdge(int from, int to)
    {
	for (int x = 0; x < nedges; x++) 
	    {
		if (((edges[x].from == from) && (edges[x].to == to)) ||
		    ((edges[x].to == from) && (edges[x].from == to))) 
		    return x;
	    }
	return -1;
    }
    
    /* Unmark nodes so search can bagin */
    void unmarkNodes()
    {
	for(int i=0;i<nnodes;i++)
	    {
		nodes[i].marked = false;
	    }
    }
    
    /* Find edge leading from one node to another */
    int findEdgeTo(String node1, String node2)
    {
	unmarkNodes();
	int from = findNode(node1);
	int to = findNode(node2);
	System.out.println("From "+node1+" to "+node2+" index "+from+" to "+to);
	if ((from == -1) || (to == -1))
	    return -1;
	int next = nodes[from].findNextHop(to);
	System.out.println("next "+next);
	return findEdge(from,next);
    }
    
    /* Add node to the panel */
    int addNode(IPAddress add, int x, int y,  
		String type, int id) 
    {
	Node n = new Node();
	n.x = x;
	n.y = y;
	n.lbl = add.toString();
	n.address = add.toLong();
	n.type = type;
	n.ID = id;
	n.externalID = nnodes;
	nodes[nnodes] = n;
	if (type.startsWith("router"))
	    n.label = "router";
	else
	    n.label = type;
	changed=true;
	if ((int)n.x + 100 > size.width)
	    {
		size.width = (int)n.x + 100;
	    }
	if ((int)n.y + 100 > size.height)
	    {
		size.height = (int)n.y + 100;
	    }
	TrackingFrame tf = new TrackingFrame(n);
	n.tf = tf;
	n.panel = tf.panel;
	System.out.println("Added node "+add+" at index "+nnodes);
	return nnodes++;
    }
    
    /* Add edge to the panel */
    void addEdge(long from, long to)
    {
	int a =  FindAddress((long)from);
	int b =  FindAddress((long)to);
	
	Edge e = new Edge(nodes[a],nodes[b]);
	e.fromnode = from;
	e.tonode = to;
	e.from = a; 
	e.to = b; 
	
	if ((e.from == -1) || (e.to == -1))
	    return;
	/* check to see if edge is already there */
	for (int x = 0; x < nedges; x++) 
	    {
		if (((edges[x].from == e.from) && (edges[x].to == e.to)) ||
		    ((edges[x].to == e.from) && (edges[x].from == e.to))) 
		    {
			return;
		    }
	    }
	
	edges[nedges++] = e;
	if (nodes[e.from].findNeighbor(nodes[e.to]) == -1)
	    nodes[e.from].addNeighbor(nodes[e.to]);
	if (nodes[e.to].findNeighbor(nodes[e.from]) == -1)
	    nodes[e.to].addNeighbor(nodes[e.from]);
	changed=true;
    } 

    /* Add sniffer to the panel */
    void addSniff(int from, int to)
    {
	Sniff s = new Sniff(edges[from],edges[to]);
	s.from = from; 
	s.to = to; 
	
	sniffs[nsniffs++] = s;
	changed=true;
    } 
    
    /* Find node given the index */
    public long FindNode(int index)
    {
	if (index>nnodes)
	    return -1;
	else
	    return nodes[index].address;
    } 

    /* Find index of the node given the address */
    public int FindAddress(long address)
    {
	for (int i=0;i<nnodes;i++)
	    {
		if (nodes[i].address==address)
		    return i;
	    }
	return -1;
    }
    
    /* Find node given  the label */
    int findNode(String lbl) 
    {
	for (int i = 0 ; i < nnodes ; i++) 
	    {
		if (nodes[i].lbl.equals(lbl)) 
		    {
			return i;
			}
	    }
	return -1;
    }
    
    /* Zoom in */
    public void zoom_in(Graphics g) 
    {
	scale *= 1.25;
	Dimension size = drawingArea.getSize();
	size.width = (int)(size.width*1.25);
	size.height = (int)(size.height*1.25);
	drawingArea.setPreferredSize(size);
	drawingArea.revalidate();
	drawingArea.repaint();
    }
    
    /* Zoom out */
    public void zoom_out(Graphics g) 
    {
	changed=true;
	scale *= 0.8;
	Dimension size = drawingArea.getSize();
	size.width = (int)(size.width*0.8);
	size.height = (int)(size.height*0.8);
	drawingArea.setPreferredSize(size);
	drawingArea.revalidate();
	drawingArea.repaint();
    }

    /* Paint the time on panel */
    public void paintTime(int time, int oldtime)
    {
	int sec = time%60;
	int min = (time/60)%60;
	int hour = time/3600;
	String tm = hour+":"+min+":"+sec;
	int seco = oldtime%60;
	int mino = (oldtime/60)%60;
	int houro = oldtime/3600;
	String to = houro+":"+mino+":"+seco;
	Graphics g = drawingArea.getGraphics();
	Dimension d = getSize();
	Font fnt = new Font("Sans Serif",Font.PLAIN,30);
	g.setFont(fnt);
	FontMetrics fm = g.getFontMetrics();
	int fw = fm.stringWidth("0");
	int hpos = d.width/2-4*fw;
	int mpos = d.width/2-fw;
	int spos = d.width/2+2*fw;
	int h = fm.getHeight()+10;
	
	g.setColor(BgColor);
	if (hour != houro)
	    {
		g.fillRect(hpos,0,2*fw,h);
	    }
	if (min!=mino)
	    {
		g.fillRect(mpos,0,2*fw,h);
	    }
	if (sec!=seco)
	    {
		g.fillRect(spos,0,2*fw,h);
	    }
	g.setColor(Color.red);
	if (hour<10)
	    g.drawString("0"+hour,hpos,h);
	else
	    g.drawString(""+hour,hpos,h);
	if (min<10)
	    g.drawString("0"+min,mpos,h);
	else
	    g.drawString(""+min,mpos,h);
	if (sec<10)
	    g.drawString("0"+sec,spos,h);
	else
	    g.drawString(""+sec,spos,h);
	g.drawString(":",(int)(hpos+2.3*fw),h);
	g.drawString(":",(int)(mpos+2.3*fw),h);
    }
}




