/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

class Edge {
    int from, to;
    long fromnode, tonode;
    int fint, tint;
        final static Color default_color = Color.black;
    Color edge_color = Color.black; 
    Color from_color = Color.gray;
    Color to_color =  Color.gray;
    
    /* Packets for the victim on the edge at every second */
    Vector packets;
    /* Replies from the victim on the edge at every second */
    Vector replies;
    /* Attack packets for the victim on the edge at every second */
    Vector apackets;
    /* Bytes for the victim on the edge at every second */
    Vector bytes;
    /* Bytes from the victim on the edge at every second */
    Vector rbytes;
    /* Attack bytes for the victim on the edge at every second */
    Vector abytes;

    int offset;
    int xint1,xint2,yint1,yint2;
    Node tnode, fnode;
    double alpha;

    public Edge(Node f, Node t)
    {
	fnode = f;
	tnode = t;
	offset=3;
	packets = new Vector();
	replies = new Vector();
	apackets = new Vector();
	bytes = new Vector();
	rbytes = new Vector();
	abytes = new Vector();
    }
    
    /* Insert packets for a given time */
    public void InsertPacket(int num, int time, int b)
    {
	if (packets.size()<time)
	    packets.setSize(time);
	packets.add(time, new Integer(num));
	if (bytes.size()<time)
	    bytes.setSize(time);
	bytes.add(time, new Integer(b));
    }

    /* Insert replies for a given time */
    public void InsertReply(int num, int time, int rb)
    {
	if (replies.size()<time)
	    replies.setSize(time);
	replies.add(time, new Integer(num));
	if (rbytes.size()<time)
	    rbytes.setSize(time);
	rbytes.add(time, new Integer(rb));
    }

    /* Insert attack packets for a given time */
    public void InsertAttackPacket(int num, int time, int b)
    {
	if (apackets.size()<time)
	    apackets.setSize(time);
	apackets.add(time, new Integer(num));
	if (abytes.size()<time)
	    abytes.setSize(time);
	abytes.add(time, new Integer(b));
    }

    /* Find a line paralel to this edge at distance d */
    Line2D.Double getParallelLine(int d)
    {
	int x1 = (int)(xint1-d*Math.sin(alpha));
	int y1 = (int)(yint1+d*Math.cos(alpha));
	int x2 = (int)(xint2-d*Math.sin(alpha));
	int y2 = (int)(yint2+d*Math.cos(alpha));
	Line2D.Double line = new Line2D.Double(x1,y1,x2,y2);
	return line;
    }

    /* Calculate intersection between the edge and its nodes */ 
    public void calculateIntersection(double scale)
    {
	int x1 = (int)(fnode.x * scale);
	int y1 = (int)(fnode.y * scale);
	int x2 = (int)(tnode.x * scale);
	int y2 = (int)(tnode.y * scale);
	
	alpha=Math.atan((double)(y2-y1)/(x2-x1));
	
	Line2D.Double line = new Line2D.Double(x1,y1,x2,y2);
	
	/* Find intersection with the first node, for every edge
	   assuming rectangle shape */
	if (line.intersectsLine(fnode.x*scale-(fnode.width/2), 
				fnode.y*scale-(fnode.height/2),
				fnode.x*scale-(fnode.width/2), 
				fnode.y*scale+(fnode.height/2)))
	    {
		xint1 =(int)(fnode.x*scale-(fnode.width/2));
		yint1 = (int)(fnode.y*scale 
			      - (fnode.width/2)*Math.tan(alpha));
	    }
	else if (line.intersectsLine(fnode.x*scale+(fnode.width/2), 
				     fnode.y*scale-(fnode.height/2),
				     fnode.x*scale+(fnode.width/2), 
				     fnode.y*scale+(fnode.height/2)))
	    {
		xint1 =(int)(fnode.x*scale + (fnode.width/2));
		yint1 = (int)(fnode.y*scale 
			      + (fnode.width/2)*Math.tan(alpha));
	    }
	else if (line.intersectsLine(fnode.x*scale-(fnode.width/2), 
				     fnode.y*scale-(fnode.height/2),
				     fnode.x*scale+(fnode.width/2), 
				     fnode.y*scale-(fnode.height/2)))
	    {
		yint1 =(int)(fnode.y*scale-(fnode.height/2));
		xint1 = (int)(fnode.x*scale 
			      - (fnode.height/2)/Math.tan(alpha));
	    }
	else
	    {
		yint1 =(int)(fnode.y*scale + (fnode.height/2));
		xint1 = (int)(fnode.x*scale 
			      + (fnode.height/2)/Math.tan(alpha));
	    }

	/* Find intersection with the second node, for every edge
	   assuming rectangle shape */
	
	if (line.intersectsLine(tnode.x*scale-(tnode.width/2), 
				tnode.y*scale-(tnode.height/2),
				tnode.x*scale-(tnode.width/2), 
				tnode.y*scale+(tnode.height/2)))
	    {
		xint2 =(int)(tnode.x*scale-(tnode.width/2));
		yint2 = (int)(tnode.y*scale 
			      - (tnode.width/2)*Math.tan(alpha));
	    }
	else if (line.intersectsLine(tnode.x*scale+(tnode.width/2), 
				     tnode.y*scale-(tnode.height/2),
				     tnode.x*scale+(tnode.width/2), 
				     tnode.y*scale+(tnode.height/2)))
	    {
		xint2 =(int)(tnode.x*scale + (tnode.width/2));
		yint2 = (int)(tnode.y*scale 
			      + (tnode.width/2)*Math.tan(alpha));
	    }
	else if (line.intersectsLine(tnode.x*scale-(tnode.width/2), 
				     tnode.y*scale-(tnode.height/2),
				     tnode.x*scale+(tnode.width/2), 
				     tnode.y*scale-(tnode.height/2)))
	    {
		yint2 =(int)(tnode.y*scale-(tnode.height/2));
		xint2 = (int)(tnode.x*scale 
			      - (tnode.height/2)/Math.tan(alpha));
	    }
	else
	    {
		yint2 =(int)(tnode.y*scale + (tnode.height/2));
		xint2 = (int)(tnode.x*scale 
			      + (tnode.height/2)/Math.tan(alpha));
	    }
    }
}








