/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/************************************************************************
 * This file contains a few utility functions for network communication *
 * (GUI receives reports through this interface)                        *
 ************************************************************************/

#include "NetworkUtil.h"
#include "server.h"
#include "constants.h"
#include "dward.config"

#include <stdio.h>
#include <sys/errno.h>
#include <pthread.h>

extern int GUIPORT;
static int socketFD = 0;
static int connfd[50];
static int conn = 0;
static int LEN = 100;
char strings[10000];

char message[10000];
int mes_pos = 0;

pthread_mutex_t count_mutex; 
pthread_t thread[100]; 

/* read_data:
 *	reads exactly n characters from the socket
 *	arguments and return value as in read
 */

int read_data(int s,char *buf,int n)
{
  int bcount, br;
  bcount = 0;
  br = 0;
  while (bcount < n)
  {
    if ((br = read(s,buf,n-bcount)) > 0)
    {
      bcount += br;
      buf += br;
    }
    if (br < 0)
      return(-1);
  }

  return(bcount);
}

/* write_data:
 *	writes exactly n items to the socket
 *	arguments and return value as in write
 */

int write_data(int s,char *buf,int n)
{
  int bcount, br;

  bcount = 0;
  br = 0;
  while (bcount < n)
  {
    if ((br = write(s, buf, n-bcount)) > 0)
    {
      bcount += br;
      buf += br;
    }
    if (br < 0)
    {
      perror("write_data()");
      return(-1);
    }
  }

  return(bcount);
}


/* establish:
 *	establishes a socket to accept incoming requests
 */

int establish(int port)
{
  int s;
  struct sockaddr_in sa;
  char myname[MAXHOSTNAME + 1];
  struct hostent *hp;

  bzero(&sa, sizeof(struct sockaddr_in));
  strcpy(myname, "localhost");
  hp = gethostbyname(myname);
  if (hp == NULL)
    return (-1);

  sa.sin_family = hp->h_addrtype;
  sa.sin_port = htons((u_short)port);
  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    return (-1);

  if (bind(s, (struct sockaddr *)&sa, sizeof sa) < 0)
    return (-1);
  listen(s, BACKLOG);
  return (s);
}


/* get_connection:
 *	accept incoming requests,
 *	return a new socket that the incoming request
 *	will send and receive data
 */

int get_connection(int s)
{
  struct sockaddr_in isa;
  int t;
  int i;

  i = sizeof(isa);
  getsockname(s, (struct sockaddr *)&isa, &i);      
  if ((t = accept(s, (struct sockaddr *)&isa,&i)) < 0)
    return(-1);
  return(t);
}

/* Receive data from open connection,
	 this is threaded function */
void handle_connection(int *sid)
{
	fd_set readfds;
	int numbytes; 
	unsigned char buf[1000];
	send(*sid,strings,strlen(strings),0);
	FD_CLR(*sid, &readfds);
	FD_SET(*sid, &readfds);  
	listen (*sid, 50);
	while(1)
		{
		  if (select((*sid)+1, &readfds, NULL, NULL, NULL) == -1) 
				{
					perror("select");
					exit(1);
				}			
			if (FD_ISSET(*sid,&readfds))
			{
				numbytes = recv(*sid, buf, LEN, 0);
				buf[numbytes] = 0;
				pthread_mutex_lock(&count_mutex);
				memcpy(message+mes_pos,buf,strlen(buf));
				mes_pos += strlen(buf);
				pthread_mutex_unlock(&count_mutex);
			}			
		}
}

/* call_socket:
 *	sets up connection with the (known) server who is
 *	listening to the port.
 *	returns -1 on failure, 0 on success
 */

int call_socket(char *host, int port)
{
  struct sockaddr_in sa, their_addr;
  socklen_t sin_size = sizeof(struct sockaddr);
  struct hostent *hp;
  int maxFD;
  fd_set readfds;
  short i;
  int on = 1;

  if ((hp = gethostbyname("localhost")) == NULL)
    return (-1);
	
  bzero(&sa, sizeof(sa));
  bcopy(hp->h_addr_list[0], (char *)&sa.sin_addr, hp->h_length);
	
  sa.sin_family = hp->h_addrtype;
  sa.sin_port = htons(GUIPORT);
  sa.sin_addr.s_addr = INADDR_ANY;
  
  if ((socketFD = socket(hp->h_addrtype, SOCK_STREAM, 0)) < 0)
    return (-1);
  setsockopt (socketFD, SOL_SOCKET, SO_REUSEADDR, 
	      (void *) &on, sizeof (on));
  setsockopt (socketFD, SOL_SOCKET, SO_REUSEPORT, 
	      (void *) &on, sizeof (on));	
  if (bind(socketFD, (struct sockaddr *)&sa, sizeof(struct sockaddr_in))   == -1) 
    {
      perror("bind");
      return 1;
    }
  
  listen (socketFD, 50);
  FD_CLR(socketFD, &readfds);
  FD_SET(socketFD, &readfds);  
  maxFD = socketFD;
  
  while(1)
    {
      if (select(maxFD+1, &readfds, NULL, NULL, NULL) == -1) 
	{
	  perror("select");
	  exit(1);
	}
      for (i=0;i<=maxFD;i++)
	{
	  /* New connection request */
	  if (FD_ISSET(i, &readfds))
	    {
	      if (i==socketFD)
		{
		  connfd[conn] = accept(socketFD, (struct sockaddr *)&their_addr, &sin_size);
		  conn++;
		  pthread_create(&thread[conn], NULL, (void*)handle_connection, &connfd[conn-1]); 
		}
	    }
	}
    }
  return 0;
}

/* Initializes message string */
void InitializeStrings(char* s)
{
	strcpy(strings,s);
}

/*
 * ConnectToServer:
 *   function that connects to server on a specified port
 */

void ConnectToServer()
{
	int thread_id = 0;
	pthread_create(&thread[0], NULL, (void*)call_socket, &thread_id); 
	return;
}

/*
 * SendRequest:
 *   function that sends a request to the server
 */

int SendRequest(Request *req, int port)
{
  int wsize,reqsize = sizeof(Request);

  if(!socketFD)
  {
    ConnectToServer(port);
    fprintf(stderr, "Opened TCP connection FDid:%d to server\n", socketFD);
  }

  if ((wsize =write_data(socketFD, (char *)req, reqsize)) != reqsize)
  {
    fprintf(stderr, "FATAL ERROR when writing to socket\n");
  }

  return socketFD;
}

/*
 * getSocketID:
 *   function that returns value of global variable socketFD
 */

int getSocketID()
{
  if(socketFD)
    return socketFD;
  else
    return -1;
}


/*
 * ReceiveReply:
 *   function that receives reply from the peer on a socket sFD
 */

int ReceiveReply(int sFD, unsigned char *reply, int replySize)
{
  int n;
  n = read_data(sFD, (unsigned char *) reply, replySize);
  if (n == -1)
  {
    fprintf(stderr, "READ_DATA FAILED\n");
    perror("ReceiveReply()");
  }
  else if (n != replySize)
  {
    fprintf(stderr, "ReceiveReply: size mismatch\n");
  }

  return 1;
}

/* Copy message into shared variable so that 
	 GUI can read it */
void GetMessage(char* buf)
{
	pthread_mutex_lock(&count_mutex);
	strcpy(buf,message);
	buf[mes_pos] = 0;
	message[0]= 0;
	mes_pos = 0;
	pthread_mutex_unlock(&count_mutex);
}

















