/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* Hash table structures and constants */
#ifndef __HASH_H
#define __HASH_H

/* Define when full alert should be raised */
#define FULL_ALERT_LF 0.9
#define NEW 1
#define EXISTING 2

/* Hash table entry */
struct hash_entry{
  /* Key used for hashing */
  void* key;
  /* Additional data */
  void* data;
  /* Flag to show state of entry (VALID, EMPTY, DELETED) */
  int state;
  /* Flag to show if entry has already been visited,
     serves to prevent infinite loop */
  long visited;
};

/* Hash table 
   This is really an attempt to
   make classfull, reusable
   structure in C */
struct hash_table{
  /* Contents */
  struct hash_entry* entries;
  /* Number of entries */
  unsigned long size;
  unsigned long min_size;
  unsigned long max_size;
  /* Number of filled entries */
  unsigned long filled;
  /* Size of key */
  int key_size;
  /* Size of data */
  int data_size;
  /* Load factors for rehashing */
  double rehash_lf_high;
  double rehash_lf_low;
  /* Flag indicating that hash is dangerously full and 
     can not grow anymore. Then drastic cleaning has to
     be done */
  int FULL_ALERT;
  /* Every insert, find and delete 
     will increment this and it should
     wrap-around when needed. This is the value
     to mark visited entry */
  long stamp;
  /* Function to check if entry can be deleted
     depending on urgency of deletion */
  int (*clean)(void*);
  /* Function to compare keys if they are same 
     it returns 1 if they are same, and 0 otherwise */
  int (*compare)(void*, void*);
  /* First hash function, for double hashing */
  unsigned long (*hash1)(void*,unsigned long);
  /* Second hash function, for double hashing */
  unsigned long (*hash2)(void*,unsigned long);
  /* Sometimes it is useful to forbid rehashing */
  int no_rehash;
};


/* Function prototypes, refer to implementation in hash.c for descriptions */
int test_prime(long);
void create_hash(struct hash_table**,  unsigned long, 
                 unsigned long, unsigned long, 
                 double, double, int, int, 
		 int (*)(void*), int (*)(void*,void*),
                 unsigned long (*)(void*,unsigned long), 
                 unsigned long (*)(void*,unsigned long));
void clean_table(struct hash_table*);
void rehash(int, struct hash_table*);
long insert(void*, struct hash_table*, int*);
long find(void*, struct hash_table*);
long delete(void*, struct hash_table*);
void check_rehash(struct hash_table*);

#endif



