/* Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

#include <linux/time.h>
#include <linux/config.h>  
#include <linux/slab.h> 
#include <linux/mm.h> 
#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/types.h> 
#include <linux/stddef.h> 
#include <linux/types.h> 
#include <linux/netfilter.h> 
#include <linux/netfilter_ipv4.h> 
#include <linux/skbuff.h> 
#include <linux/in.h> 
#include <linux/ip.h> 
#include <linux/tcp.h> 
#include <linux/udp.h> 
#include <linux/icmp.h> 
#include <linux/netdevice.h> 
#include <linux/if_ether.h> 
#include <sys/syscall.h>  
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/types.h> 
#include <linux/netfilter.h> 
#include <linux/netfilter_ipv4.h> 
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <asm/semaphore.h>
#include <endian.h>
#include <unistd.h>
#include <linux/devfs_fs_kernel.h>

#include "machines.h"
#include "constants.h"

/* Transform 32-bit network address 
   into dotted notation */
void ntoa(__u32 address, char* temp)
{
  __u32 host_order = ntohl(address);
  int pos = 0;
  int div = 16777216;
  while(div)
    {
      int n = host_order/div;
      host_order = host_order % div;
      if (div>1)
	pos += sprintf(temp+pos,"%d.",n);
      else
	pos += sprintf(temp+pos,"%d",n);
      div = div/256;
    }
}

int find_port(struct machine* m, int p)
{
  int i;
  for(i=0;i<MAX_DATA;i++)
    if(m->port[i] == p)
      return i;
  return NOT_FOUND;
}

void delete_port(struct machine* m, int p)
{
  int i;
  for(i=0;i<MAX_DATA;i++)
    if(m->port[i] == p)
    {
      m->port[i] = 0;
      m->num_data--;
      return;
    }
}

int insert_data(struct machine* m, __u16 port, __u32 seqnum)
{
 int i;
 for(i=0;i<MAX_DATA;i++)
   /* If found an empty slot, insert data */
   if(m->port[i] == 0)
     {
       struct timeval tv;
       do_gettimeofday(&tv);
       m->port[i] = port;
       m->seqnum[i] = seqnum;
       m->time[i] = tv.tv_sec;
       return i;
     }
 return NOT_FOUND;
}

int validate_data(struct machine* m, int index)
{
  if (m->cur != index)
    {
      m->prev = m->cur;
      m->cur = index;
      m->num_data++;
    }
  return index;
}

void init_machine(struct machine* m)
{
  int i;
  for(i=0;i<MAX_DATA;i++)
    m->port[i] = 0;
  m->num_data = 0;
  m->os_type = NOS;
}

int classify_machine(struct machine* m)
{
  if (m->num_data >= 2)
    {
      __u32 prediction;
      /* First assume system is Windows */
      prediction = predict_seqnum(m, m->prev, m->time[m->cur], WINDOWS);
      if (verify_prediction(prediction, m->seqnum[m->cur],WINDOWS))
	{
	  char temp[17];
	  m->os_type = WINDOWS;
	  ntoa(m->address, temp);
	  return 1;
	}
      /* Then assume it is Linux */
      prediction = predict_seqnum(m, m->prev, m->time[m->cur],LINUX);
      if (verify_prediction(prediction, m->seqnum[m->cur],LINUX))
	{
	  char temp[17];
	  m->os_type = LINUX;
	  ntoa(m->address, temp);
	  return 1;
	}
      /* Else it is FreeBSD */
      m->os_type = FREEBSD;
      return 1;
    }
  return 0;
}

int verify_prediction(__u32 prediction, __u32 real_value, enum os os_type)
{
  __u32 em, diff;
  if(os_type == WINDOWS)
    em = (__u32)WIN_ERROR_MARGIN;
  else
    em = (__u32)LIN_ERROR_MARGIN;
  if (real_value > prediction)
    diff = real_value - prediction;
  else
    diff = prediction - real_value;
  /* Check for wraparound */
  if (((MAX_SEQNUM - prediction < em) && (real_value < em)) ||
      ((MAX_SEQNUM - real_value < em) && (prediction < em)))
  {
    diff = abs(MAX_SEQNUM - diff);
  }
  //printk("diff is %u\n",diff);
  if(diff < em)
    return 1;
  return 0;
}

__u32 predict_seqnum(struct machine* m, int index, int time, enum os os_type)
{
  __u32 prediction;
  if (os_type == WINDOWS)
    {
      prediction = m->seqnum[index] + 
	(time - m->time[index])*WIN_FACTOR 
      + WIN_CONSTANT;
      if (prediction > MAX_SEQNUM)
	prediction = WIN_CONSTANT;
      return prediction;
    }
  else if(os_type == LINUX)
    {
      prediction = m->seqnum[index] + 
	(time-m->time[index])*LIN_FACTOR 
	+ LIN_CONSTANT;
      if (prediction > MAX_SEQNUM)
	prediction = LIN_CONSTANT;
      return prediction;
    }
  else 
    return 0;
}





