/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* System includes */
#include <linux/time.h> 
#include <linux/config.h>  
#include <linux/slab.h> 
#include <linux/mm.h> 
#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/types.h> 
#include <linux/stddef.h> 
#include <linux/types.h> 
#include <linux/netfilter.h> 
#include <linux/netfilter_ipv4.h> 
#include <linux/skbuff.h> 
#include <linux/in.h> 
#include <linux/ip.h> 
#include <linux/tcp.h> 
#include <linux/udp.h> 
#include <linux/icmp.h> 
#include <linux/netdevice.h> 
#include <linux/if_ether.h> 
#include <sys/syscall.h>  
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/types.h> 
#include <linux/netfilter.h> 
#include <linux/netfilter_ipv4.h> 
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <asm/semaphore.h>
#include <endian.h>
#include <unistd.h>
#include <linux/devfs_fs_kernel.h>

#include "sniff.h"

/* Command line initialization parameter */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
MODULE_PARM(plen, "i");
#endif

struct semaphore s;

extern struct file_operations sniff_fops;

/* Buffer to hold packet info */
unsigned char* buffer; 
/* How many packers are in buffer */
int packet_count = 0;
int old_packet_count;
int idle = 0;

/* Packet len */
int plen = 40;
int full = 0;

void read_buffer(unsigned char* outbuffer, int* pl, int* pc)
{
  /* down(&s); */
  if (packet_count > BUF_LEN/(10*plen) || idle > 1000)
    {
      memcpy(outbuffer,buffer,packet_count*plen);
      *pl=plen;
      *pc=packet_count;
      packet_count = 0;
      idle = 0;
      full=0;
    }
  else
    {
      if (packet_count == old_packet_count)
	idle ++;
      else
	idle = 0;
      old_packet_count = packet_count;
      *pl=plen;
      *pc=0;
    }
  /* up(&s); */
}

/* Filtering function, called whenever a packet is intercepted */
static unsigned int nf_packet_passing(unsigned int hook, struct sk_buff **pskb,
				      const struct net_device *indev, 
				      const struct net_device *outdev, 
				      int (*okfn)(struct sk_buff *)) 
{ 
    if ((packet_count+1)*plen < BUF_LEN)
    {
      struct iphdr *ip=(*pskb)->nh.iph;
      if (plen <= ntohs(ip->tot_len))
	{
	  /* down(&s); */
	  memcpy(buffer+packet_count*plen,ip,plen);
	  packet_count++;
	  /* up(&s); */
	}
      else /* Pad with zeros */
	{
	  memcpy(buffer+packet_count*plen,ip,ntohs(ip->tot_len));
	  memset(buffer+packet_count*plen+ntohs(ip->tot_len),0,plen-ntohs(ip->tot_len));
	  packet_count++;
	}
    }
  else
    {
      if (!full)
	printk("Full after %d packets\n",packet_count);
      full=1;
      /* Should wrap around here */
    }
    return NF_ACCEPT;
}

static struct nf_hook_ops forward_chain = { { NULL, NULL }, nf_packet_passing, PF_INET, NF_IP_FORWARD, 0 }; 

/* Initialization function, called when a module is loaded */
int init_module()
{
  init_MUTEX(&s);
  /* Allocate buffer */
  buffer = kmalloc(BUF_LEN, GFP_KERNEL);
  if (buffer == NULL)
    {
      printk("Unable to allocate %d bytes of memory \n",BUF_LEN);
      return -EFAULT;
    }
  /* Open SNIFF device */
  if (devfs_register_chrdev (SNIFF_MAJOR, "sniff", &sniff_fops)) {
    printk ("sniff: unable to get major %d\n", SNIFF_MAJOR);
    return -EIO;
  }
  nf_register_hook(&forward_chain);
  return 0;
}


/* Restore function, called when a module is unloaded */
void cleanup_module()
{
  kfree(buffer);
  devfs_unregister_chrdev(SNIFF_MAJOR, "sniff");
  nf_unregister_hook(&forward_chain); 
}










