#!/usr/bin/python2
#
# D-WARD statistics analyzer
# Written by Adam Vartanian, September 4, 2003
#
# Reads D-WARD parsable statistics from stdin or file
# Outputs information on stdout

import sys, getopt, time

def usage(e=None):
    if e != None:
        print >>sys.stderr,e
    print >>sys.stderr,"Usage: analyze_bytes [OPTIONS] [INPUT FILE]"
    print >>sys.stderr,"Reads input from INPUT FILE, or standard input if no input file given."
    print >>sys.stderr,"Input is assumed to be a properly formatted D-WARD traffic stats file."
    print >>sys.stderr,"The results from providing any other kind of input are undefined."
    print >>sys.stderr,"Options:"
    print >>sys.stderr,"  --help              Print this message and exit"
    print >>sys.stderr,"  -c, --continuous    Print continuous information to stdout"
    print >>sys.stderr,"  -b, --beep          Beep whenever packets are dropped"
    print >>sys.stderr,"  -a N, --alarm N     Every N seconds, display general and current ratelimit information"
    sys.exit(1)

def printstats():
    firsttime = time.mktime(time.strptime(firstdate))
    lasttime = time.mktime(time.strptime(lastdate))
    elapsed = lasttime - firsttime
    timeunit = "seconds"
    if elapsed > 60:
        elapsed /= 60
        timeunit = "minutes"
        if elapsed > 60:
            elapsed /= 60
            timeunit = "hours"
            if elapsed > 36:
                elapsed /= 24
                timeunit = "days"

    totalbytes = generalstats['forwarded'] + generalstats['dropped']
    fw = float(generalstats['forwarded'])
    fwsuffix = ""
    if fw > 1024:
        fw /= 1024
        fwsuffix = "K"
        if fw > 1024:
            fw /= 1024
            fwsuffix = "M"
            if fw > 1024:
                fw /= 1024
                fwsuffix = "G"
    dp = float(generalstats['dropped'])
    dpsuffix = ""
    if dp > 1024:
        dp /= 1024
        dpsuffix = "K"
        if dp > 1024:
            dp /= 1024
            dpsuffix = "M"
            if dp > 1024:
                dp /= 1024
                dpsuffix = "G"
    
    print "General Summary:"
    print "  Data from %s to %s (%.2f %s)" % (firstdate, lastdate, elapsed, timeunit)
    print "%10.2f%s forwarded (%0.4f%%)" % (fw,fwsuffix, 100*float(generalstats['forwarded'])/totalbytes)
    print "%10.2f%s dropped (%0.4f%%)" % (dp,dpsuffix, 100*float(generalstats['dropped'])/totalbytes)


try:
    opts, pargs = getopt.getopt(sys.argv[1:], 'bca:', ['beep','continuous','alarm=','help'])
except getopt.GetoptError,e:
    usage(e)

if len(pargs) > 1:
    usage('analyze_stats: too many arguments')

beep = 0
general = 0
continuous = 0
alarmlength = 0
firstdate = None
lastdate = None
generalstats = { 'forwarded' : 0,
                 'dropped' : 0,
                 'errors' : [] }

for i in opts:
    if i[0] == '-b' or i[0] == '--beep':
        beep = 1
    elif i[0] == '-c' or i[0] == '--continuous':
        continuous = 1
    elif i[0] == '-a' or i[0] == '--alarm':
        alarmlength = int(i[1])
    elif i[0] == '--help':
        usage()

if len(pargs) == 1:
    try:
        sys.stdin = open(pargs[0])
    except IOError,e:
        print e
        sys.exit(2)

atime = time.time()
try:
    while 1:
        if (time.time() - alarmlength) > atime and alarmlength > 0:
            printstats()
            atime = time.time()
        line = raw_input()
        linelist = line.replace('  ',' ').split(' ')
        for i in range(len(linelist)):
            linelist[i] = linelist[i].strip()
        if firstdate == None:
            firstdate = line[:24]
        lastdate = line[:24]
        generalstats['forwarded'] += int(linelist[6])
        generalstats['dropped'] += int(linelist[5])
        if int(linelist[5]) > 0 and continuous and beep:
            print "\a%s: Dropped %d bytes" % (line[:24], int(linelist[5]))
        elif int(linelist[5]) > 0 and continuous:
            print "%s: Dropped %d bytes" % (line[:24], int(linelist[5]))
        elif int(linelist[5]) > 0 and beep:
            print "\a",
                
except (KeyboardInterrupt, EOFError),e:
    pass

for i in generalstats['errors']:
    print "Error: %s" % i
print ""

printstats()
