#!/usr/bin/python2
#
# D-WARD statistics analyzer
# Written by Adam Vartanian, September 4, 2003
#
# Reads D-WARD parsable statistics from stdin or file
# Outputs information on stdout

import sys, getopt, time, signal

def usage(e=None):
    if e != None:
        print >>sys.stderr,e
    print >>sys.stderr,"Usage: analyze_stats [OPTIONS] [INPUT FILE]"
    print >>sys.stderr,"Reads input from INPUT FILE, or standard input if no input file given."
    print >>sys.stderr,"Input is assumed to be a properly formatted parsable D-WARD stats file."
    print >>sys.stderr,"The results from providing any other kind of input are undefined."
    print >>sys.stderr,"Options:"
    print >>sys.stderr,"  --help              Print this message and exit"
    print >>sys.stderr,"  -c, --continuous    Print continuous information to stdout"
    print >>sys.stderr,"  -b, --beep          Beep whenever a rate limit is imposed"
    print >>sys.stderr,"  -g, --general       Print a general summary of all input when finished"
    print >>sys.stderr,"  -d, --dest          Print a destination summary of all input when finished"
    print >>sys.stderr,"  -s, --source        Print a source summary of all input when finished"
    print >>sys.stderr,"  -p, --proto         Print a protocol summary of all input when finished"
    print >>sys.stderr,"  -a N, --alarm N     Every N seconds, display general and current ratelimit information"
    sys.exit(1)

def printstats(pgen = 0, pproto = 0, psrc = 0, pdest = 0, pcurrl = 0):
    if pgen == 1:
        firsttime = time.mktime(time.strptime(firstdate))
        lasttime = time.mktime(time.strptime(lastdate))
        elapsed = lasttime - firsttime
        timeunit = "seconds"
        if elapsed > 60:
            elapsed /= 60
            timeunit = "minutes"
            if elapsed > 60:
                elapsed /= 60
                timeunit = "hours"
                if elapsed > 36:
                    elapsed /= 24
                    timeunit = "days"

        print "General Summary:"
        print "  Data from %s to %s (%.2f %s)" % (firstdate, lastdate, elapsed, timeunit)
        print "%6d good connections" % generalstats['good']
        print "%6d bad connections" % generalstats['bad']
        print "%6d ratelimits imposed" % generalstats['ratelimit']
        print "%6d seconds of ratelimiting" % generalstats['ratelimittime']
        if len(generalstats['classifications'].keys()) > 0:
            print "  Issued the following classifications:"
            for i in generalstats['classifications'].keys():
                print "%25s: %d times" % (i, generalstats['classifications'][i][0])
        print ""

    if pproto == 1:
        print "Protocol Summary:"
        for i in protocolstats.keys():
            j = protocolstats[i]
            print "  %5s: %d total; %d good (%0.2f%%), %d bad (%0.2f%%)" % (i, j[0], j[1], 100*float(j[1])/j[0], j[2], 100*float(j[2])/j[0])
        print ""

    if psrc == 1:
        print "Source Summary:"
        for i in sourcestats.keys():
            j = sourcestats[i]
            print "  %15s: %d total; %d good (%0.2f%%), %d bad (%0.2f%%)" % (i, j[0], j[1], 100*float(j[1])/j[0], j[2], 100*float(j[2])/j[0])
        print ""

    if pdest == 1:
        print "Destination Summary:"
        for i in deststats.keys():
            j = deststats[i]
            print "  %15s: %d total; %d good (%0.2f%%), %d bad (%0.2f%%)" % (i, j[0], j[1], 100*float(j[1])/j[0], j[2], 100*float(j[2])/j[0])
        print ""

    if pcurrl == 1:
        if len(currentratelimits.keys()) == 0:
            print "No active ratelimits"
        else:
            print "Current ratelimits:"
            for i in currentratelimits.keys():
                j = currentratelimits[i]
                print "  %15s at %s (started at %s)" % (i, j[1], j[0])


try:
    opts, pargs = getopt.getopt(sys.argv[1:], 'sbcdgpa:', ['source','dest','proto','beep','general','continuous','protocol','destination','alarm=','help'])
except getopt.GetoptError,e:
    usage(e)

if len(pargs) > 1:
    usage('analyze_stats: too many arguments')

beep = 0
general = 0
dests = 0
sources = 0
protocols = 0
continuous = 0
alarmlength = 0
firstdate = None
lastdate = None
generalstats = { 'good' : 0,
                 'bad' : 0,
                 'ratelimit' : 0,
                 'ratelimittime' : 0,
                 'classifications' : {},
                 'errors' : [] }
currentratelimits = {}
sourcestats = {}
deststats = {}
protocolstats= {}

for i in opts:
    if i[0] == '-s' or i[0] == '--source':
        sources = 1
    elif i[0] == '-b' or i[0] == '--beep':
        beep = 1
    elif i[0] == '-d' or i[0] == '--dest' or i[0] == '--destination':
        dests = 1
    elif i[0] == '-g' or i[0] == '--general':
        general = 1
    elif i[0] == '-p' or i[0] == '--proto' or i[0] == '--protocol':
        protocols = 1
    elif i[0] == '-c' or i[0] == '--continuous':
        continuous = 1
    elif i[0] == '-a' or i[0] == '--alarm':
        alarmlength = int(i[1])
    elif i[0] == '--help':
        usage()

if len(pargs) == 1:
    try:
        sys.stdin = open(pargs[0])
    except IOError,e:
        print e
        sys.exit(2)

atime = time.time()
try:
    while 1:
        if (time.time() - alarmlength) > atime and alarmlength > 0:
            printstats(1,0,0,0,1)
            atime = time.time()
        line = raw_input()
        linelist = line.split('\t')
        for i in range(len(linelist)):
            linelist[i] = linelist[i].strip()
        if firstdate == None:
            firstdate = linelist[2]
        lastdate = linelist[2]
        if linelist[0] == "C":
            if linelist[1] == "C":
                if protocolstats.has_key(linelist[8]):
                    protocolstats[linelist[8]][0] += 1
                else:
                    protocolstats[linelist[8]] = [1,0,0]
                if sourcestats.has_key(linelist[3]):
                    sourcestats[linelist[3]][0] += 1
                else:
                    sourcestats[linelist[3]] = [1,0,0]
                if deststats.has_key(linelist[5]):
                    deststats[linelist[5]][0] += 1
                else:
                    deststats[linelist[5]] = [1,0,0]

                if linelist[9] == 'GOOD':
                    generalstats['good'] += 1
                    protocolstats[linelist[8]][1] += 1
                    sourcestats[linelist[3]][1] += 1
                    deststats[linelist[5]][1] += 1
                elif linelist[9] == 'BAD':
                    generalstats['bad'] += 1
                    protocolstats[linelist[8]][2] += 1
                    sourcestats[linelist[3]][2] += 1
                    deststats[linelist[5]][2] += 1
            elif linelist[1] == "F":
                if generalstats['classifications'].has_key(linelist[5]):
                    generalstats['classifications'][linelist[5]][0] += 1
                else:
                    generalstats['classifications'][linelist[5]] = [1]
        elif linelist[0] == "RL":
            if beep:
                print '\a',
            if linelist[1] == "DW-ED":
                if not currentratelimits.has_key(linelist[3]):
                    currentratelimits[linelist[3]] = [linelist[2], linelist[5]]
                    if continuous == 1:
                        print "New ratelimit on %s, allowing %s through" % (linelist[3], linelist[5])
                else:
                    currentratelimits[linelist[3]][1] = linelist[5]
                    if continuous == 1:
                        print "Ratelimit on %s decreased to %s" % (linelist[3], linelist[5])
            elif linelist[1] == "DW-EI" or linelist[1] == "DW-MAX" or linelist[1] == "DW-LI":
                if not currentratelimits.has_key(linelist[3]):
                    generalstats['errors'].append("%s: D-WARD increased ratelimit on unratelimited stream?" % linelist[2])
                    currentratelimits[linelist[3]] = [linelist[2], linelist[5]]
                    if continuous == 1:
                        print "New ratelimit on %s, allowing %s through" % (linelist[3], linelist[5])
                else:
                    currentratelimits[linelist[3]][1] = linelist[5]
                    if continuous == 1:
                        print "Ratelimit on %s increased to %s" % (linelist[3], linelist[5])
            elif linelist[1] == "DW-REM":
                if not currentratelimits.has_key(linelist[3]):
                    generalstats['errors'].append("%s: D-WARD removed ratelimit on unratelimited stream?" % linelist[2])
                    if continuous == 1:
                        print "Unknown ratelimit on %s removed" % linelist[3]
                else:
                    rlstart = time.mktime(time.strptime(currentratelimits[linelist[3]][0]))
                    rlend = time.mktime(time.strptime(linelist[2]))
                    generalstats['ratelimit'] += 1
                    generalstats['ratelimittime'] += rlend - rlstart
                    del currentratelimits[linelist[3]]
                    if continuous == 1:
                        print "Ratelimit on %s removed after %d seconds" % (linelist[3], rlend - rlstart)            

except (KeyboardInterrupt, EOFError),e:
    pass

for i in generalstats['errors']:
    print "Error: %s" % i
print ""

printstats(general, protocols, sources, dests, 0)
